#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "typedef.h"
#include <math.h>
#include <float.h>

#include "codec.h"

#define N_DUMMY_INT 0

static int n_nbitflag = 0;
static unsigned int n_b2s[32];

N_SI4 n_ipow(const N_SI4 i, const N_SI4 j)
{
    N_SI4 k, rt;
    for (k = 0, rt = 1; k < j; k++)
	rt *= i;
    return (rt);
}

double dpow2( int num )
{
    double x = 1.0;
    while ( num-- ) x *= 2.0;
    return x;
}

static int n_nbit_init(const int nbit){
  N_UI4 wi;
  int i;

  if (n_nbitflag != nbit)  {
    wi = 1;
    for (i = 0; i < nbit; i++){
      wi *= 2;
    }
    wi = wi - 1;

    for(i = 0; i < 32; i++) {
      n_b2s[i] = wi << i;
    }
    n_nbitflag = nbit;
  }
  return 0;
}

int n_nbit_unpack(unsigned char din[], int nin, unsigned int dout[], 
                  int nout, int nbit)
/* -------------------------------------------------------------------------- */
{
   
  unsigned int wi;
  int i, bitp, dp;
   
  n_nbit_init(nbit);
   
  i = 0; dp = 0; bitp = 0;
  while(dp < nin)  {
    bitp += nbit;
    wi = 0;
    if(dp < nin - 3){
        wi = (((unsigned int)(din[dp])) << 24)
            | (((unsigned int)(din[dp+1])) << 16)
            | (((unsigned int)(din[dp+2])) << 8)
            | (((unsigned int)(din[dp+3])));
    }
    else if(dp == nin - 3){
        if(bitp > 24){
            return (i);
        }
        wi= (((unsigned int)(din[dp])) << 24)
            | (((unsigned int)(din[dp+1])) << 16)
            | (((unsigned int)(din[dp+2])) << 8);
    }
    else if(dp == nin - 2){
        if(bitp > 16){
            return (i);
        }
        wi= (((unsigned int)(din[dp])) << 24)
            | (((unsigned int)(din[dp+1])) << 16);
    }
    else if(dp == nin - 1){
        if(bitp > 8){
            return (i);
        }
        wi= (((unsigned int)(din[dp])) << 24);
    }
    dout[i++] = (wi >> (32 - bitp)) & n_b2s[0];
    while (bitp >= 8) {
      dp++;
      bitp -= 8;
    }
    if (i > nout)
      return(-1);
  }
  
  return(i);
}

int n_nbit_pack(N_UI4 din[], N_SI4 nin, N_UI1 dout[], N_SI4 nout, N_SI4 nbit)
{
   
  N_UI4 wi;
  N_UI1 *wout;
  int i, j, bitp, dp;
   

  n_nbit_init(nbit);
   
  for(i = 0; i < nout; i++)
    dout[i] = 0;
   
  for(i = 0, bitp = 0, dp = 0, wout = &dout[dp];
              i < nin; i++, wout = &dout[dp]) {
    bitp += nbit;
    wi = ((unsigned int)din[i] << (32 - bitp)) & n_b2s[32 - bitp];
    for(j=0;j<4;j++){
        *(wout+j) |= ((wi >> (24-8*j)) & 0xff);
    }
    while (bitp >= 8) {
      dp++;
      bitp -= 8;
    }
    if ((dp + 3) > nout)
      return(-1);
  }
  if (bitp != 0)
    dp++;
   
  return(dp);
}


N_SI4 
n_decode_rlen_nbit(N_UI1 *udata, N_UI1 *din, N_SI4 nin, N_SI4 nout,
                            N_SI4 maxv, N_SI4 nbit)
/* -------------------------------------------------------------------------- */
/* decode run length packed streams ( nbit data version )                     */
/*   output:   dout   = user data for put                                     */
/*   input:    din    = compressed data                                       */
/*             nin    = compressed data size (byte)                           */
/*             nout   = allocated item number for user data                   */
/*             maxv   = maximum value of user data                            */
/*             nbit   = number of bit used for a compressed data              */
/*   return:   >=0    = number of decoded data                                */
/*             -4     = uncompressed data size exceeds nout                   */
/*             -6     = first user data is out of the data range              */
/* -------------------------------------------------------------------------- */
{
  N_SI4 i, j, k, l, m = N_DUMMY_INT, n = N_DUMMY_INT, v, p, *dout, ninb;
  N_UI4 *wdat;
   
  wdat = (N_UI4 *)malloc(sizeof(N_UI4) * nout);
  ninb = n_nbit_unpack(din, nin, wdat, nout, nbit);
  if (ninb < 0)
    return(-4);
  dout = (N_SI4 *)udata;
  l = n_ipow(2, nbit) - 1 - maxv;
  v = (N_SI4)(*wdat);
  if (v < 0 || v > maxv)
    return(-6);
  
  i = 0; k = 0; p = -1;
  while(i < ninb) {
    v = (N_SI4)(*(wdat + i++));
    if (v <= maxv) {
      if (p >= 0) {
	for(j = 0; j < m; j++) {
	  if (k == nout)
	    return(-4);
	  *(dout + k++) = p;
	}
      }
      p = v;
      m = 1;
      n = 0;
    } else {
      m += n_ipow(l, n++) * (v - maxv - 1);
    }
  }
  for(j = 0; j < m; j++) {
    if (k == nout)
      return(-4);
    *(dout + k++) = p;
  }
   
  free(wdat);
  return(k);
}




