#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "gpv.h"

void gpv_write_init(grds_t *grds, grib2_t *grib2, gpv_t *gpv){
  gpv->grds = grds;
  gpv->grib2 = grib2;
  return;
}

int gpv_write_pre(const N_UI1 jma_kind, 
		  const char *fname_out, const grid_info_t *gi_out,
		  const N_UI1 pdef_use, 
		  const N_UI1 earth_shape,
		  const N_UI4 nz, double *layer_out,
		  const N_UI1 layer_kind, const ymdhm_t *ymdhm, 
		  const N_UI2 cutoff_time_hour, const N_UI1 cutoff_time_min, 
		  const N_SI4 term,
		  grib2_t *grib2_in, gpv_t *gpv){
  int r;

  gpv->nz = nz;
  gpv->layer = layer_out;
  gpv->layer_kind = layer_kind;
  gpv->ymdhm = *ymdhm;
  gpv->cutoff_time_hour = cutoff_time_hour;
  gpv->cutoff_time_min = cutoff_time_min;
  gpv->term = term;

  if (gpv->grds != NULL){
    switch(jma_kind){
    case (JMA_KIND_ATM):
      grds_def(fname_out, 
	       gi_out, pdef_use, 
	       nz, layer_out, layer_kind,
	       ymdhm, gpv->grds);
      break;
    case (JMA_KIND_OCEAN):
      grds_def_sst(fname_out, 
		   gi_out, pdef_use, 
		   ymdhm, gpv->grds);
      break;
    case (JMA_KIND_LAND):
      grds_def_land(fname_out, 
		    gi_out, pdef_use, 
		    nz, layer_out, 
		    ymdhm, gpv->grds);
      break;
    case (JMA_KIND_RA):
      grds_def_ra(fname_out, 
		  gi_out,pdef_use, 
		  ymdhm, gpv->grds);
      break;
    }

    if ((r = grds_file_open(gpv->grds))) return r;
    if ((r = grds_ctl_write(gpv->grds))) return r;


  } else if (gpv->grib2 != NULL){
    grib2_init(gpv->grib2);
    
    gpv->grib2->data.list = 
      (grib2_data_list_t*)malloc(sizeof(grib2_data_list_t));

    gpv->grib2->data.list->sec0 = (sec0_t*)malloc(sizeof(sec0_t));
    gpv->grib2->data.list->sec1 = (sec1_t*)malloc(sizeof(sec1_t));
    gpv->grib2->data.list->sec2 = (sec2_t*)malloc(sizeof(sec2_t));
    gpv->grib2->data.list->sec3 = (sec3_t*)malloc(sizeof(sec3_t));
    gpv->grib2->data.list->sec4 = (sec4_t*)malloc(sizeof(sec4_t));
    gpv->grib2->data.list->sec5 = (sec5_t*)malloc(sizeof(sec5_t));
    gpv->grib2->data.list->sec6 = (sec6_t*)malloc(sizeof(sec6_t));
    gpv->grib2->data.list->sec7 = (sec7_t*)malloc(sizeof(sec7_t));

    if ((r = grib2_open(fname_out, "w", gpv->grib2))) return r;

    if ((r = grib2_write_sec0123(earth_shape, 
				 gi_out, grib2_in, gpv->grib2))) return r;

  } else {
    fprintf(stderr, "A valid file format is not specified.\n");
    return ERR_GPV_INVALID_FORMAT;
  }
  return 0;
}


int gpv_write_run(const char *elem, N_UI4 nlevels_in, fdat_t *fdat, gpv_t *gpv){
  N_UI4 nlevels;
  N_UI1 layer_kind;

  if (gpv->grds != NULL){
    grds_dat_write(elem, nlevels_in, fdat, gpv->grds);
  } else if (gpv->grib2 != NULL){
    if (nlevels_in == 0){
      if (memcmp(elem, "PSEA", 4) == 0){
	layer_kind = LAYER_MSL;
      } else {
	layer_kind = LAYER_SURFACE;
      }
      nlevels = 1;
    } else {
      layer_kind = gpv->layer_kind;
      nlevels = nlevels_in;
    }
    
    grib2_write_sec4567(elem, 
			gpv->cutoff_time_hour, 
			gpv->cutoff_time_min,
			gpv->term,
			layer_kind, gpv->layer, 
			nlevels, 
			&(gpv->ymdhm), 
			fdat->data, 
			gpv->grib2);
  }
  return 0;
}

void gpv_write_post(gpv_t *gpv){
  if (gpv->grds != NULL){
    grds_file_close(gpv->grds);
  } else if (gpv->grib2 != NULL){
    grib2_write_end(gpv->grib2);
    grib2_close(gpv->grib2);

    free(gpv->grib2->data.list->sec0);
    free(gpv->grib2->data.list->sec1);
    free(gpv->grib2->data.list->sec2);
    free(gpv->grib2->data.list->sec3);
    free(gpv->grib2->data.list->sec4);
    free(gpv->grib2->data.list->sec5);
    free(gpv->grib2->data.list->sec6);
    free(gpv->grib2->data.list->sec7);

    free(gpv->grib2->data.list);
  }

  return;
}

int gpv_write_convert(const grid_info_t *gi_org, const int nz, 
                     const ymdhm_t *ymdhm, const N_SI4 term, 
                     double *org_layer_list, z2p_info_t *z2p,
                     grid_info_t *gi_out, interp_2d_t *interp,
                     grib2_t *grib2_in, gpv_t *gpv){
    const double r_pref = 1.0 / PREF;
    const double rdvcp = RD / CP;

    N_UI1 parm_cat;
    N_UI1 parm_num;

    fdat_t press_org;

    fdat_t data_org;
    fdat_t data_v;
    fdat_t data_h;

    fdat_t data_u_org;
    fdat_t data_u_v;
    fdat_t data_u_h;

    fdat_t data_v_org;
    fdat_t data_v_v;
    fdat_t data_v_h;


    /* for isopressure */
    fdat_t pt_org;
    fdat_t t_org;
    fdat_t z_org;
    fdat_t psurf_org;
    fdat_t zs_org;
    fdat_t tsurf_org;
    fdat_t t_v;


    fdat_t *data_wrt;
    fdat_t *data_u_wrt;
    fdat_t *data_v_wrt;

    double layer_surface = 0.0;  /* dummy */

    unsigned int nlv_wrt;
    unsigned int k;
    unsigned int ixy;
    int r;


    fdat_init(&press_org);

    fdat_init(&data_org);
    fdat_init(&data_v);
    fdat_init(&data_h);

    fdat_init(&data_u_org);
    fdat_init(&data_u_v);
    fdat_init(&data_u_h);

    fdat_init(&data_v_org);
    fdat_init(&data_v_v);
    fdat_init(&data_v_h);


    if (z2p != NULL){
      fdat_prepare(gi_org->nxy * nz, &press_org);
      grib2_get_parm("P     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  press_org.data))) return r;
      
      if ((r = z2p_pre(nz, press_org.data, z2p))) return r;
    }

    /* zs */
    fdat_prepare(gi_org->nxy, &data_org);
    grib2_get_parm("Z     ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			 ymdhm, term, 
			 parm_cat, parm_num, 
			 LAYER_SURFACE, 1, &layer_surface, 
			data_org.data))) return r;

    data_wrt = &data_org;
    if (interp != NULL){
      fdat_prepare(gi_out->nxy, &data_h);
      interp_run(gi_org, gi_out, 1, interp, 
                  data_wrt->data, data_h.data);
      data_wrt = &data_h;
    }
    gpv_write_run("ZS", 0, data_wrt, gpv);

    /* Psurf */
    fdat_prepare(gi_org->nxy, &data_org);
    grib2_get_parm("P     ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_SURFACE, 1, &layer_surface, 
			data_org.data))) return r;
    data_wrt = &data_org;
    if (interp != NULL){
      fdat_prepare(gi_out->nxy, &data_h);
      interp_run(gi_org, gi_out, 1, interp, 
                  data_org.data, data_h.data);
      data_wrt = &data_h;
    }
    gpv_write_run("Psurf", 0, data_wrt, gpv);

    /* PSEA */
    fdat_prepare(gi_org->nxy, &data_org);
    grib2_get_parm("PSEA  ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_MSL, 1, &layer_surface, 
			data_org.data))) return r;
    data_wrt = &data_org;
    if (interp != NULL){
      fdat_prepare(gi_out->nxy, &data_h);
      interp_run(gi_org, gi_out, 1, interp, 
                  data_org.data, data_h.data);
      data_wrt = &data_h;
    }
    gpv_write_run("MSLP", 0, data_wrt, gpv);

    /* U */
    fdat_prepare(gi_org->nxy * nz, &data_u_org);
    grib2_get_parm("U     ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_u_org.data))) return r;
    if (z2p == NULL){
      data_u_wrt = &data_u_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_u_v);
      z2p_gen_run(z2p, data_u_org.data, data_u_v.data);
      data_u_wrt = &data_u_v;
      nlv_wrt = z2p->pnum;
    }

    /* V */
    fdat_prepare(gi_org->nxy * nz, &data_v_org);
    grib2_get_parm("V     ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_v_org.data))) return r;
    if (z2p == NULL){
      data_v_wrt = &data_v_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v_v);
      z2p_gen_run(z2p, data_v_org.data, data_v_v.data);
      data_v_wrt = &data_v_v;
      nlv_wrt = z2p->pnum;
    }

    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_u_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                  data_u_wrt->data, data_u_h.data);
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_v_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                  data_v_wrt->data, data_v_h.data);
      interp_wndrot(gi_org, gi_out, nlv_wrt, 
                    interp->fi, interp->fj, 
                    data_u_h.data, data_v_h.data);
      data_u_wrt = &data_u_h;
      data_v_wrt = &data_v_h;
    }
    gpv_write_run("U", nlv_wrt, data_u_wrt, gpv);
    gpv_write_run("V", nlv_wrt, data_v_wrt, gpv);

    fdat_free(&data_u_org);
    fdat_free(&data_u_v);
    fdat_free(&data_u_h);
    fdat_free(&data_v_org);
    fdat_free(&data_v_v);
    fdat_free(&data_v_h);

    /* W */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("W     ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                  data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("W", nlv_wrt, data_wrt, gpv);

    if (z2p == NULL){
      nlv_wrt = nz;
      /* Z */
      fdat_prepare(gi_org->nxy * nz, &data_org);
      grib2_get_parm("Z     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  data_org.data))) return r;
      data_wrt = &data_org;
      if (interp != NULL){
        fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
        interp_run(gi_org, gi_out, nlv_wrt, interp, 
                    data_wrt->data, data_h.data);
        data_wrt = &data_h;
      } 
      gpv_write_run("Z", nlv_wrt, data_wrt, gpv);
      
      /* P */
      fdat_prepare(gi_org->nxy * nz, &data_org);
      grib2_get_parm("P     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  data_org.data))) return r;
      data_wrt = &data_org;
      if (interp != NULL){
        fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
        interp_run(gi_org, gi_out, nlv_wrt, interp, 
                    data_wrt->data, data_h.data);
        data_wrt = &data_h;
      } 
      gpv_write_run("P", nlv_wrt, data_wrt, gpv);

      /* PT */
      fdat_prepare(gi_org->nxy * nz, &data_org);
      grib2_get_parm("PT    ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  data_org.data))) return r;
      data_wrt = &data_org;
      if (interp != NULL){
        fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
        interp_run(gi_org, gi_out, nlv_wrt, interp, 
                    data_wrt->data, data_h.data);
        data_wrt = &data_h;
      } 
      gpv_write_run("PT", nlv_wrt, data_wrt, gpv);
    } else {
      fdat_init(&pt_org);
      fdat_init(&t_org);
      fdat_init(&z_org);
      fdat_init(&psurf_org);
      fdat_init(&zs_org);
      fdat_init(&tsurf_org);
      fdat_init(&t_v);

      /* read PT */
      fdat_prepare(gi_org->nxy * nz, &pt_org);
      grib2_get_parm("PT    ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  pt_org.data))) return r;

      /* read Z */
      fdat_prepare(gi_org->nxy * nz, &z_org);
      grib2_get_parm("Z     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_HYBRID, nz, org_layer_list, 
			  z_org.data))) return r;

      /* read Psurf */
      fdat_prepare(gi_org->nxy, &psurf_org);
      grib2_get_parm("P     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_SURFACE, 1, &layer_surface, 
			  psurf_org.data))) return r;

      fdat_prepare(gi_org->nxy, &zs_org);
      grib2_get_parm("Z     ", &parm_cat, &parm_num);
      if ((r = grib2_read(grib2_in, 
			  ymdhm, term, 
			  parm_cat, parm_num, 
			  LAYER_SURFACE, 1, &layer_surface, 
			  zs_org.data))) return r;

      /* convert PT to T */
      fdat_prepare(gi_org->nxy * nz, &t_org);
      for (k = 0; k < nz; k++){
        for (ixy = 0; ixy < gi_org->nxy; ixy++){
          t_org.data[k * gi_org->nxy + ixy] = 
            pt_org.data[k * gi_org->nxy + ixy] 
            * pow(press_org.data[k * gi_org->nxy + ixy] * r_pref, rdvcp);
        }
      }
      fdat_prepare(gi_org->nxy, &tsurf_org);
      for (ixy = 0; ixy < gi_org->nxy; ixy++){
	tsurf_org.data[ixy] = pt_org.data[0 * gi_org->nxy + ixy] 
                    * pow(psurf_org.data[ixy] * r_pref, rdvcp);
      }
      fdat_free(&pt_org);


      fdat_prepare(gi_org->nxy * z2p->pnum, &t_v);
      z2p_temp_run(z2p, 
                   psurf_org.data, 
                   t_org.data, t_v.data);

      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gph_run(z2p, zs_org.data, z_org.data, 
                  tsurf_org.data,  
                  t_v.data, 
                  data_v.data);

      fdat_free(&z_org);
      fdat_free(&psurf_org);
      fdat_free(&zs_org);
      fdat_free(&tsurf_org);
      fdat_free(&t_org);

      /* write */
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
      if (interp != NULL){
        fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
        interp_run(gi_org, gi_out, nlv_wrt, interp, 
                    data_wrt->data, data_h.data);
        data_wrt = &data_h;
      } 
      gpv_write_run("Z", nlv_wrt, data_wrt, gpv);

      data_wrt = &t_v;
      nlv_wrt = z2p->pnum;
      if (interp != NULL){
        fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
        interp_run(gi_org, gi_out, nlv_wrt, interp, 
                    data_wrt->data, data_h.data);
        data_wrt = &data_h;
      } 
      gpv_write_run("T", nlv_wrt, data_wrt, gpv);


      fdat_free(&t_v);
    }

    /* QV */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("QV    ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QV", nlv_wrt, data_wrt, gpv);

    /* QC */
    grib2_get_parm("QC    ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QC", nlv_wrt, data_wrt, gpv);

    /* QR */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("QR    ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QR", nlv_wrt, data_wrt, gpv);


    /* QCI */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("QCI   ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QCI", nlv_wrt, data_wrt, gpv);

    
    /* QS */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("QS    ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QS", nlv_wrt, data_wrt, gpv);


    /* QG */
    fdat_prepare(gi_org->nxy * nz, &data_org);
    grib2_get_parm("QG    ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_HYBRID, nz, org_layer_list, 
			data_org.data))) return r;
    if (z2p == NULL){
      data_wrt = &data_org;
      nlv_wrt = nz;
    } else {
      fdat_prepare(gi_org->nxy * z2p->pnum, &data_v);
      z2p_gen_run(z2p, data_org.data, data_v.data);
      data_wrt = &data_v;
      nlv_wrt = z2p->pnum;
    }
    if (interp != NULL){
      fdat_prepare(gi_out->nxy * nlv_wrt, &data_h);
      interp_run(gi_org, gi_out, nlv_wrt, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("QG", nlv_wrt, data_wrt, gpv);
    
    /* RAIN */
    fdat_prepare(gi_org->nxy, &data_org);
    grib2_get_parm("RAIN  ", &parm_cat, &parm_num);
    if ((r = grib2_read(grib2_in, 
			ymdhm, term, 
			parm_cat, parm_num, 
			LAYER_SURFACE, 1, &layer_surface, 
			data_org.data))) return r;
    data_wrt = &data_org;
    if (interp != NULL){
      fdat_prepare(gi_out->nxy, &data_h);
      interp_run(gi_org, gi_out, 1, interp, 
                 data_wrt->data, data_h.data);
      data_wrt = &data_h;
    } 
    gpv_write_run("RAIN", 0, data_wrt, gpv);


    fdat_free(&press_org);

    fdat_free(&data_org);
    fdat_free(&data_v);
    fdat_free(&data_h);
    return 0;
}


int gpv_write_convert_sst(const grid_info_t *gi_org,  
			 const ymdhm_t *ymdhm, const N_SI4 term, 
			 grid_info_t *gi_out, interp_2d_t *interp, 
			 grib2_t *grib2_in, gpv_t *gpv){

  fdat_t data_org;
  fdat_t data_h;
  fdat_t *data_wrt;

  N_UI1 parm_cat;
  N_UI1 parm_num;
  
  double layer_surface = 0.0;  /* dummy */
  int r;

  fdat_init(&data_org);
  fdat_init(&data_h);


  fdat_prepare(gi_org->nxy, &data_org);
  grib2_get_parm("SST   ", &parm_cat, &parm_num);
  if ((r = grib2_read(grib2_in, 
		      ymdhm, term, 
		      parm_cat, parm_num, 
		      LAYER_SURFACE, 1, &layer_surface, 
		      data_org.data))) return r;
  data_wrt = &data_org;
  if (interp != NULL){
    fdat_prepare(gi_out->nxy, &data_h);
    interp_run(gi_org, gi_out, 1, interp, 
	       data_wrt->data, data_h.data);
    data_wrt = &data_h;
  }
  gpv_write_run("SST", 0, data_wrt, gpv);

  fdat_free(&data_org);
  fdat_free(&data_h);

  return 0;

}

int gpv_write_convert_land(const grid_info_t *gi_org, const int nz, 
			  const ymdhm_t *ymdhm, const N_SI4 term, 
			  double *org_layer_list,
			  grid_info_t *gi_out, interp_2d_t *interp, 
			  grib2_t *grib2_in, gpv_t *gpv){
  fdat_t data_org;
  fdat_t data_h;
  fdat_t *data_wrt;

  N_UI1 parm_cat;
  N_UI1 parm_num;
  
  double layer_surface = 0.0;
  int r;

  fdat_init(&data_org);
  fdat_init(&data_h);


  /* TUGD */
  fdat_prepare(gi_org->nxy * nz, &data_org);
  grib2_get_parm("TUGD  ", &parm_cat, &parm_num);
  if ((r = grib2_read(grib2_in, 
		 ymdhm, term, 
		 parm_cat, parm_num, 
		 LAYER_UGD, nz, org_layer_list, 
		      data_org.data))) return r;
  data_wrt = &data_org;
  if (interp != NULL){
    fdat_prepare(gi_out->nxy * nz, &data_h);
    interp_run(gi_org, gi_out, nz, interp, 
	       data_wrt->data, data_h.data);
    data_wrt = &data_h;
  }
  gpv_write_run("TUGD", nz, data_wrt, gpv);

  /* KIND */
  fdat_prepare(gi_org->nxy, &data_org);
  grib2_get_parm("KIND  ", &parm_cat, &parm_num);
  if ((r = grib2_read(grib2_in, 
		      ymdhm, term, 
		      parm_cat, parm_num, 
		      LAYER_SURFACE, 1, &layer_surface, 
		      data_org.data))) return r;
  data_wrt = &data_org;
  if (interp != NULL){
    fdat_prepare(gi_out->nxy, &data_h);
    interp_nearest_run(gi_org, gi_out, 1, interp, 
	       data_wrt->data, data_h.data);
    data_wrt = &data_h;
  }
  gpv_write_run("KIND", 0, data_wrt, gpv);


  fdat_free(&data_org);
  fdat_free(&data_h);

  return 0;


}

int gpv_write_convert_ra(const grid_info_t *gi_org,  
			const ymdhm_t *ymdhm, const N_SI4 term, 
			const N_UI1 ra_interp_ope, 
			grid_info_t *gi_out, interp_2d_t *interp,
			grib2_t *grib2_in, gpv_t *gpv){

  fdat_t data_org;
  fdat_t data_h;
  fdat_t *data_wrt;

  N_UI1 parm_cat;
  N_UI1 parm_num;
  
  double layer_surface = 0.0;  /* dummy */
  int r;

  fdat_init(&data_org);
  fdat_init(&data_h);


  fdat_prepare(gi_org->nxy, &data_org);
  grib2_get_parm("RR60  ", &parm_cat, &parm_num);
  if ((r = grib2_read(grib2_in, 
		      ymdhm, term, 
		      parm_cat, parm_num, 
		      LAYER_SURFACE, 1, &layer_surface, 
		      data_org.data))) return r;
  data_wrt = &data_org;
  if (interp != NULL){
    fdat_prepare(gi_out->nxy, &data_h);
    switch (ra_interp_ope){
    case (INTERP_OPE_NEAR):
      interp_nearest_run(gi_org, gi_out, 1, interp, 
			 data_wrt->data, data_h.data);
      break;
    case (INTERP_OPE_MEAN):
    case (INTERP_OPE_MAX):
      interp_mxme(gi_org, gi_out, 
		  ra_interp_ope, 
		  data_wrt->data, data_h.data);
      break;
    }

    data_wrt = &data_h;
  }
  gpv_write_run("RAIN", 0, data_wrt, gpv);

  fdat_free(&data_org);
  fdat_free(&data_h);

  return 0;

}

