#ifndef GPV_H_INCLUDED
#define GPV_H_INCLUDED

#include "grib2_io.h"
#include "grads_io.h"
#include "typedef.h"
#include "ymdhm.h"
#include "grads_wrt.h"
#include "grib2_encode.h"
#include "z2p.h"
#include "interp.h"

typedef struct gpv_t{
  grds_t *grds;
  grib2_t *grib2;

  N_UI4 nz;
  double *layer;
  N_UI1 layer_kind;
  ymdhm_t ymdhm;
  N_UI2 cutoff_time_hour;
  N_UI1 cutoff_time_min;
  N_SI4 term;
} gpv_t;

void gpv_write_init(grds_t *grds, grib2_t *grib2, gpv_t *gpv);
int gpv_write_pre(const N_UI1 jma_kind, 
		  const char *fname_out, const grid_info_t *gi_out,
		  const N_UI1 pdef_use,
		  const N_UI1 earth_shape, 
		  const N_UI4 nz, double *layer_out,
		  const N_UI1 kind_zcord, const ymdhm_t *ymdhm, 
		  const N_UI2 cutoff_time_hour, const N_UI1 cutoff_time_min, 
		  const N_SI4 term,
		  grib2_t *grib2_in, gpv_t *gpv);
int gpv_write_run(const char *elem, N_UI4 nlevels_in, fdat_t *fdat, gpv_t *gpv);
void gpv_write_post(gpv_t *gpv);
int gpv_write_convert(const grid_info_t *gi_org, const int nz, 
		      const ymdhm_t *ymdhm, const N_SI4 term, 
		      double *org_layer_list, z2p_info_t *z2p,
		      grid_info_t *gi_out, interp_2d_t *interp,
		      grib2_t *grib2, gpv_t *gpv);

int gpv_write_convert_sst(const grid_info_t *gi_org,  
			  const ymdhm_t *ymdhm, const N_SI4 term, 
			  grid_info_t *gi_out, interp_2d_t *interp, 
			  grib2_t *grib2_in, gpv_t *gpv);

int gpv_write_convert_land(const grid_info_t *gi_org, const int nz, 
			   const ymdhm_t *ymdhm, const N_SI4 term, 
			   double *org_layer_list,
			   grid_info_t *gi_out, interp_2d_t *interp, 
			   grib2_t *grib2_in, gpv_t *gpv);

int gpv_write_convert_ra(const grid_info_t *gi_org,  
			 const ymdhm_t *ymdhm, const N_SI4 term, 
			 const N_UI1 ra_interp_ope, 
			 grid_info_t *gi_out, interp_2d_t *interp,
			 grib2_t *grib2_in, gpv_t *gpv);

#endif
