#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "endian.h"
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "endian.h"
#include "grads_io.h"

const char month_str[][4]= {"Jan", "Feb", "Mar", "Apr", "May", "Jun", 
                            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

grds_opt_list_t *grds_opt_list_add(const char *option, grds_t *grds){

  grds_opt_list_t *item_add;

  item_add = (grds_opt_list_t*)malloc(sizeof(grds_opt_list_t));  
 
  item_add->option = (char*)malloc(strlen(option) + 1);
  strcpy(item_add->option, option);
  item_add->next = NULL;

  if (grds->opt_list == NULL){
    grds->opt_list = item_add;
  } else {
    grds->opt_last->next = item_add;
  }
  grds->opt_last = item_add;

  return item_add;
}

grds_var_list_t *grds_var_list_add(const char *name, const unsigned int nlevels, 
				 const char *details, 
				 grds_t *grds){
  grds_var_list_t *item_add;

  item_add = (grds_var_list_t*)malloc(sizeof(grds_var_list_t));

  item_add->grds_var = (grds_var_t*)malloc(sizeof(grds_var_t));

  item_add->grds_var->name = (char*)malloc(strlen(name) + 1);
  item_add->grds_var->details = (char*)malloc(strlen(details) + 1);
  
  strcpy(item_add->grds_var->name, name);
  item_add->grds_var->nlevels = nlevels;
  strcpy(item_add->grds_var->details, details);  
  item_add->next = NULL;

  if (grds->var_list == NULL){
    grds->var_list = item_add;
  } else {
    grds->var_last->next = item_add;
  }
  grds->var_last = item_add;
  grds->nvars++;
  return item_add;
}

void grds_meta_set(const char* str, const int meta_kind, grds_t *grds){
  char *p_target;


  p_target = (char*)malloc(strlen(str) + 1);
  strcpy(p_target, str);

  switch(meta_kind){
  case(GRDS_META_DSET):
    if (grds->dset != NULL){
      free(grds->dset);
      grds->dset = NULL;
    }
    grds->dset = p_target;
    break;
  case (GRDS_META_TITLE):
    if (grds->title != NULL){
      free(grds->title);
      grds->title = NULL;
    }
    grds->title = p_target;
    break;
  }
  return;
}

void grds_undef_set(const float undef, grds_t *grds){
  grds->undef = undef;
  return;
}

void grds_gi_set(const grid_info_t *gi, unsigned char pdef_use, grds_t *grds){
  memcpy(&(grds->gi), gi, sizeof(grid_info_t));
  grds->pdef_use = pdef_use;
  return;
}

int grds_zgrid_set(const int zgrid_kind, const unsigned int nlevels, 
		   const double *level, grds_t *grds){

  grds_zgrid_t *zgrid = &(grds->zdef);

  zgrid->zgrid_kind = zgrid_kind;
  zgrid->nlevels = nlevels;

  zgrid->level = (double*)malloc(sizeof(double) * nlevels);
  memcpy(zgrid->level, level, sizeof(double) * nlevels);

  return 0;
}

void grds_tdef_set(const unsigned int ntimes, const ymdhm_t ini_time, 
		  const unsigned int delt, grds_t *grds){
  grds_tdef_t *tdef = &(grds->tdef);

  tdef->ntimes = ntimes;
  memcpy(&(tdef->ini_time), &ini_time, sizeof(ymdhm_t));
  tdef->delt = delt;

  return;
}


void grds_init(grds_t *grds){
  grds->dset = NULL;
  grds->title = NULL;
  grds->opt_list = NULL;
  grds->nvars = 0;
  grds->var_list = NULL;
  grds->pvar = NULL;
  
  grds->opt_last = NULL;
  grds->var_last = NULL;

  grds->pdef_use = 0;

  return;
}

int grds_file_open(grds_t *grds){
  char fname_dat[256];
  char fname_ctl[256];

  strcpy(fname_dat, grds->dset);
  strcat(fname_dat, ".dat");

  strcpy(fname_ctl, grds->dset);
  strcat(fname_ctl, ".ctl");

  if ((grds->fp_dat = fopen(fname_dat, "w")) == NULL || 
      (grds->fp_ctl = fopen(fname_ctl, "w")) == NULL){
    fprintf(stderr, "Cannot open files to write\n");
    return ERR_IO_OPEN;
  }
  return 0;
}

void grds_file_close(grds_t *grds){
  fclose(grds->fp_dat);
  fclose(grds->fp_ctl);
  printf("Output to %s.ctl\n", grds->dset);
  return;
}

void grds_print_var(grds_t *grds){
  grds_var_list_t *p;

  for (p = grds->var_list; p != NULL; p = p->next){
    fprintf(stderr, "%s, %d, %s\n", 
	    p->grds_var->name, p->grds_var->nlevels, p->grds_var->details);
  }
  return;
}

int grds_dat_write(const char *elem, const unsigned int nlevels, 
	       const fdat_t *fdat, grds_t *grds){

  unsigned int nx = grds->gi.nx;
  unsigned int ny = grds->gi.ny;
  unsigned int nxy = nx * ny;
  grds_var_t *grds_var;
  unsigned int rec_size;
  int kk;
  int ii;
  int levs;
  float value;
  unsigned int ix;
  unsigned int jy;
  unsigned int idx;

  static fdat_t fdat_wrt;

  if (grds->pvar == NULL){
    grds->pvar = grds->var_list;
  }

  grds_var = grds->pvar->grds_var;
#ifdef DEBUG
  fprintf(stderr, "%s, %s, %d, %d, %d, %d\n", elem, grds_var->name, 
	  grds_var->nlevels, nlevels, nxy, fdat->esize);
#endif

  if (strlen(elem) == strlen(grds_var->name) && 
      memcmp(elem, grds_var->name, strlen(elem)) == 0 &&
      nlevels == grds_var->nlevels){

    fprintf(stderr, "Processing var: %s, nlevels: %d\n", 
	    elem, nlevels);

    
    fdat_prepare(nxy + 2 * sizeof(unsigned int),  &fdat_wrt);
    rec_size = nxy * sizeof(float);

    levs = nlevels == 0 ? 1 : nlevels;
    for (kk = 0; kk < levs; kk++){
      /* header */
      memcpy(fdat_wrt.data, &rec_size, sizeof(unsigned int));
      /* data */
      for (ii = 0; ii < nxy; ii++){
	value = *(fdat->data + kk * nxy + ii);
	ix = ii % grds->gi.nx;
	jy = ii / grds->gi.nx;
	idx = ix + ((grds->gi.ny - jy - 1) * grds->gi.nx);
	if (value == MISSING){
	  *(fdat_wrt.data + 1 + idx) = GRDS_UNDEF;
	} else {
	  *(fdat_wrt.data + 1 + idx) = value;
	}
      }
      /* footer */
      memcpy(fdat_wrt.data + 1 + nxy, &rec_size, sizeof(unsigned int));

      endian_swab4(fdat_wrt.data, nxy + 2);

      fwrite(fdat_wrt.data, 1, nxy * sizeof(float) + 2 * sizeof(unsigned int), 
	     grds->fp_dat);
    }
    grds->pvar = grds->pvar->next;

  } else {
    fprintf(stderr, "Inconsistent grd_dat_write call with pre-registered elements.\n");
    fprintf(stderr, "  %s, %s, %d, %d, %d, %d\n", elem, grds_var->name, 
	    grds_var->nlevels, nlevels, nxy, fdat->esize);

    return ERR_GRDS_IO_INCONSISTENT_WRITE;
  }
  return 0;

}

int grds_ctl_write(const grds_t *grds){
  
  char *dset_fname;
  char *option_str;
  grds_opt_list_t *p;
  grds_var_list_t *pv;
  unsigned int pos;
  unsigned int len;
  unsigned int opt_len;
  unsigned int ii;
  char *fname_base;
  char *pc;

  fname_base = grds->dset;
  pc = fname_base;

  /* fname_base points to the final '/' */
  while(*pc){
    if (*pc == '/'){
      fname_base = pc + 1;
    }
    pc++;
  }

  dset_fname = (char*)malloc(strlen(fname_base) + 4 + 1 + 1); 
            /* 4 = .dat, 1 = ^ */
  
  *dset_fname = '^';
  strcpy(dset_fname + 1, fname_base);
  strcat(dset_fname, ".dat");

  fprintf(grds->fp_ctl, "dset %s\n", dset_fname);
  fprintf(grds->fp_ctl, "title %s\n", grds->title);

  opt_len = 0;
  for(p = grds->opt_list; p != NULL; p = p->next){
    opt_len += strlen(p->option) + 1;
  }
  option_str = (char*)malloc(opt_len + 1);
  *option_str = '\0';
  pos = 0;
  for(p = grds->opt_list; p != NULL; p = p->next){
    len = strlen(p->option);
    memcpy(option_str + pos, p->option, len);
    pos += len;
    
    if (p->next != NULL){
      option_str[pos++] = ' ';
    }
  }
  option_str[pos++] = '\0';
  
  fprintf(grds->fp_ctl, "OPTIONS %s\n", option_str);
  fprintf(grds->fp_ctl, "undef %e\n", grds->undef);

  if (memcmp(grds->gi.proj, "LM", 2) == 0){
    ddat_t flat;
    ddat_t flon;
    ddat_t fi;
    ddat_t fj;
    unsigned int edge_size;
    int i;
    int j;
    double lat_max;
    double lon_max;
    double lat_min;
    double lon_min;
    double lat_start;
    double lat_end;
    double lon_start;
    double lon_end;
    double earth_rad;
    double dx;
    double dy;
    unsigned int nx;
    unsigned int ny;
    char cout_str[2];
    char none_str[1];
    char *hstr_pdef;
    char *hstr_lin;

    ddat_init(&flat);
    ddat_init(&flon);
    ddat_init(&fi);
    ddat_init(&fj);

    edge_size = (grds->gi.nx + grds->gi.ny) * 2;
    ddat_prepare(edge_size, &flat);
    ddat_prepare(edge_size, &flon);
    ddat_prepare(edge_size, &fi);
    ddat_prepare(edge_size, &fj);

    for (j = 0; j < grds->gi.ny; j++){
      fj.data[j] = (double)(j + 1);
      fj.data[j + grds->gi.ny] = (double)(j + 1);
      
      fi.data[j] = 1.0;
      fi.data[j + grds->gi.ny] = (double)(grds->gi.nx);
    }
    for (i = 0; i < grds->gi.nx; i++){
      fj.data[i + grds->gi.ny * 2] = 1.0;
      fj.data[i + grds->gi.ny * 2 + grds->gi.nx] = (double)(grds->gi.ny);
      
      fi.data[i + grds->gi.ny * 2] = (double)(i + 1);
      fi.data[i + grds->gi.ny * 2 + grds->gi.nx] = (double)(i + 1);
    }
    NWP_lambert2sphere_D(fi.data, fj.data, 
			 edge_size, grds->gi.slat1, grds->gi.slat2, 
			 grds->gi.slon, grds->gi.xlat, grds->gi.xlon, 
			 grds->gi.xi, grds->gi.xj, grds->gi.dx,
			 flat.data, flon.data);

    lat_max = -9999.0;
    lon_max = -9999.0;
    lat_min = 9999.0;
    lon_min = 9999.0;
    for (i = 0; i < edge_size; i++){
      lat_max = lat_max < flat.data[i] ? flat.data[i] : lat_max;
      lon_max = lon_max < flon.data[i] ? flon.data[i] : lon_max;
      lat_min = lat_min > flat.data[i] ? flat.data[i] : lat_min;
      lon_min = lon_min > flon.data[i] ? flon.data[i] : lon_min;
    }

    lat_start = (double)(floor(lat_min));
    lon_start = (double)(floor(lon_min));
    lat_end = (double)(floor(lat_max)) + 1.0;
    lon_end = (double)(floor(lon_max)) + 1.0;
      
    earth_rad = nwp_get_earth_rad_m();
    dx =  grds->gi.dx 
      / (earth_rad * cos(30.0 * PI / 180.0) * 2.0 * PI) * 360.0;
    dy =  grds->gi.dy / (earth_rad * 2.0 * PI) * 360.0;

    nx =  floor((lon_end - lon_start)) / dx + 1;
    ny =  floor((lat_end - lat_start)) / dy + 1;

    strcpy(cout_str, "#");
    *none_str = '\0';

    if (grds->pdef_use == 0){
      hstr_pdef = cout_str;
      hstr_lin = none_str;
    } else {
      hstr_pdef = none_str;
      hstr_lin = cout_str;
    }
    
    fprintf(grds->fp_ctl, "%spdef %d %d LCCR %f %f %f %f %f %f %f %f %f\n", 
	    hstr_pdef,
	    grds->gi.nx, grds->gi.ny, 
	    grds->gi.xlat, grds->gi.xlon, 
	    grds->gi.xi, (double)(grds->gi.ny) - grds->gi.xj + 1.0, 
	    grds->gi.slat1, grds->gi.slat2,
	    grds->gi.slon, grds->gi.dx, grds->gi.dy);
    fprintf(grds->fp_ctl, "%sxdef %d linear %f %f\n", 
	    hstr_pdef, nx, lon_start, dx);
    fprintf(grds->fp_ctl, "%sydef %d linear %f %f\n", 
	    hstr_pdef, ny, lat_start, dy);

    fprintf(grds->fp_ctl, "%sxdef %d linear %f %f\n", 
	    hstr_lin, grds->gi.nx, 1.0, 1.0);
    fprintf(grds->fp_ctl, "%sydef %d linear %f %f\n", 
	    hstr_lin, grds->gi.ny, 1.0, 1.0);

    ddat_free(&flat);
    ddat_free(&flon);
    ddat_free(&fi);
    ddat_free(&fj);
  } else if (memcmp(grds->gi.proj, "LL", 2) == 0){
    fprintf(grds->fp_ctl, "xdef %d linear %f %f\n", 
	    grds->gi.nx, grds->gi.xlon - (grds->gi.xi - 1.0) * grds->gi.dx, 
	    grds->gi.dx);
    fprintf(grds->fp_ctl, "ydef %d linear %f %f\n", 
	    grds->gi.ny, grds->gi.xlat 
            - ((double)(grds->gi.ny) - grds->gi.xj) * grds->gi.dy, 
	    grds->gi.dy);
  } else {

  }
  switch (grds->zdef.zgrid_kind){
  case (GRDS_ZGRID_LINEAR):
    fprintf(grds->fp_ctl, "zdef  1 linear 1 1\n");
    break;
  case (GRDS_ZGRID_LEVELS):
    fprintf(grds->fp_ctl, "zdef  %d levels ", grds->zdef.nlevels);    
    for (ii = 0; ii < grds->zdef.nlevels; ii++){
      fprintf(grds->fp_ctl, "%f ", grds->zdef.level[ii]);
    }
    fprintf(grds->fp_ctl, "\n");
    break;
  }

  if (grds->tdef.ntimes == 1){
    fprintf(grds->fp_ctl, "tdef  1  linear  %02d:%02dz%02d%3s%04d 1mn\n",
            grds->tdef.ini_time.hour, grds->tdef.ini_time.min, 
	    grds->tdef.ini_time.day, 
            month_str[grds->tdef.ini_time.month - 1], grds->tdef.ini_time.year);
  } else {
    fprintf(grds->fp_ctl, "tdef  %d  linear  %02d:%02dz%02d%3s%04d %dmn\n",
            grds->tdef.ntimes, 
            grds->tdef.ini_time.hour, grds->tdef.ini_time.min, 
	    grds->tdef.ini_time.day, 
            month_str[grds->tdef.ini_time.month - 1], 
            grds->tdef.ini_time.year, grds->tdef.delt);
  }
  fprintf(grds->fp_ctl, "vars  %d\n", grds->nvars);
  for (pv = grds->var_list; pv != NULL; pv = pv->next){
    fprintf(grds->fp_ctl, "%-10s %4d 99 %s\n", 
            pv->grds_var->name, pv->grds_var->nlevels, pv->grds_var->details);
  }
  fprintf(grds->fp_ctl, "endvars\n");

  free(option_str);
  free(dset_fname);
  return 0;
}
