#ifndef GRADS_IO_H
#define GRADS_IO_H

#define GRDS_ZGRID_LINEAR 0
#define GRDS_ZGRID_LEVELS 1

#define GRDS_META_DSET 0
#define GRDS_META_TITLE 1

#include "fdat.h"
#include "ddat.h"
#include "ymdhm.h"
#include "grid_info.h"
#include "nwpl_map_earth.h"
#include "nwpl_mapproj_df.h"
#include "const.h"
#include <math.h>



typedef struct grds_zgrid_t{
  int zgrid_kind;
  unsigned int nlevels;
  double *level;
} grds_zgrid_t;

typedef struct grds_tdef_t{
  unsigned int ntimes;
  ymdhm_t ini_time;
  unsigned int delt;  /* unit: min */
} grds_tdef_t;


typedef struct grds_var_t{
  char *name;
  unsigned int nlevels;
  char *details;
} grds_var_t;

typedef struct grds_var_list_t grds_var_list_t;

struct grds_var_list_t{
  grds_var_t *grds_var;
  grds_var_list_t *next;
};

typedef struct grds_opt_list_t grds_opt_list_t;

struct grds_opt_list_t{
  char *option;
  grds_opt_list_t *next;
};


typedef struct grds_t{
  char *dset;
  char *title;
  grds_opt_list_t *opt_list;
  float undef;
  grid_info_t gi;

  unsigned char pdef_use;

  grds_zgrid_t zdef;
  grds_tdef_t tdef;
  unsigned int nvars;
  grds_var_list_t *var_list;
  grds_var_list_t *pvar;
  FILE *fp_ctl;
  FILE *fp_dat;

  grds_opt_list_t *opt_last;
  grds_var_list_t *var_last;
} grds_t;


/* prototype */
grds_opt_list_t *grds_opt_list_add(const char *option, grds_t *grds);
grds_var_list_t *grds_var_list_add(const char *name, 
                                   const unsigned int nlevels, 
                                   const char *details, 
                                   grds_t *grds);
void grds_meta_set(const char* str, const int meta_kind, grds_t *grds);
void grds_undef_set(const float undef, grds_t *grds);
void grds_gi_set(const grid_info_t *gi, const unsigned char pdef_use, 
		 grds_t *grds);

int grds_zgrid_set(const int zgrid_kind, const unsigned int nlevels, 
		   const double *level, grds_t *grds);

void grds_tdef_set(const unsigned int ntimes, const ymdhm_t ini_time, 
		  const unsigned int delt, grds_t *grds);

void grds_init(grds_t *grds);
int grds_file_open(grds_t *grds);
void grds_file_close(grds_t *grds);

int grds_dat_write(const char *elem, const unsigned int nlevels, 
		   const fdat_t *fdat, grds_t *grds);
int grds_ctl_write(const grds_t *grds);

#endif
