#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "grads_wrt.h"



int grds_def(const char *fname_grib2, 
	     const grid_info_t *gi_out, const unsigned char pdef_use,
	     const unsigned int nz, const double *layer, const int kind_zcord,
	     const ymdhm_t *ymdhm, grds_t *grds){
  
  grds_init(grds);

  grds_meta_set(fname_grib2, GRDS_META_DSET, grds);
  grds_meta_set("JMA Meso Analysis converted from GRIB2", 
		GRDS_META_TITLE, grds);
  
  grds_undef_set(GRDS_UNDEF, grds);
  grds_gi_set(gi_out, pdef_use, grds);

  grds_zgrid_set(GRDS_ZGRID_LEVELS, nz, layer, grds);

  grds_tdef_set(1, *ymdhm, 0, grds);

  grds_opt_list_add("big_endian", grds);
  grds_opt_list_add("sequential", grds);

  grds_var_list_add("ZS", 0, "Topography (m)", grds);
  grds_var_list_add("Psurf", 0, "Surface Pressure (Pa)", grds);
  grds_var_list_add("MSLP", 0, "Mean sea level Pressure (Pa)", grds);
  grds_var_list_add("U", nz, "x-direction wind velocity  (m/s)", grds);
  grds_var_list_add("V", nz, "y-direction wind velocity  (m/s)", grds);
  grds_var_list_add("W", nz, "Vertical wind velocity  (m/s)", grds);
  grds_var_list_add("Z", nz, "Height (m)", grds);
  switch (kind_zcord){
  case(LAYER_HYBRID):
    grds_var_list_add("P", nz, "Pressure (Pa)", grds);    
    grds_var_list_add("PT", nz, "Potential Temperature  (K)", grds);    
    break;
  case (LAYER_ISOBAR):
    grds_var_list_add("T", nz, "Temperature (K)", grds);    
    break;
  }


  grds_var_list_add("QV", nz, "Specific humidity  (kg/kg)", grds);
  grds_var_list_add("QC", nz, "Cloud water mixing ratio  (kg/kg)", grds);
  grds_var_list_add("QR", nz, "Rain water mixing ratio  (kg/kg)", grds);
  grds_var_list_add("QCI", nz, "Cloud ice mixing ratio  (kg/kg)", grds);
  grds_var_list_add("QS", nz, "Snow mixing ratio (kg/kg)", grds);
  grds_var_list_add("QG", nz, "Graupel mixing ratio (kg/kg)", grds);
  grds_var_list_add("RAIN", 0, "3-hr accumulated precipitation (kg/m2)", 
		    grds);

#ifdef DEBUG
    grds_print_var(&grds);
#endif
    return 0;
}

int grds_def_sst(const char *fname_grib2, 
		 const grid_info_t *gi_out, const unsigned char pdef_use,
		 const ymdhm_t *ymdhm, grds_t *grds){
  
  double layer = 0.0;

  grds_init(grds);

  grds_meta_set(fname_grib2, GRDS_META_DSET, grds);
  grds_meta_set("JMA Meso Analysis (SST) converted from GRIB2", 
		GRDS_META_TITLE, grds);
  
  grds_undef_set(GRDS_UNDEF, grds);
  grds_gi_set(gi_out, pdef_use, grds);

  grds_zgrid_set(GRDS_ZGRID_LEVELS, 1, &layer, grds);

  grds_tdef_set(1, *ymdhm, 0, grds);

  grds_opt_list_add("big_endian", grds);
  grds_opt_list_add("sequential", grds);

  grds_var_list_add("SST", 0, "Sea Surface Temperature (K)", grds);

#ifdef DEBUG
    grds_print_var(&grds);
#endif
    return 0;
}

int grds_def_land(const char *fname_grib2, 
		  const grid_info_t *gi_out, const unsigned char pdef_use,
		  const unsigned int nz, const double *layer, 
		  const ymdhm_t *ymdhm, grds_t *grds){
  

  grds_init(grds);

  grds_meta_set(fname_grib2, GRDS_META_DSET, grds);
  grds_meta_set("JMA Meso Analysis (LAND) converted from GRIB2", 
		GRDS_META_TITLE, grds);
  
  grds_undef_set(GRDS_UNDEF, grds);
  grds_gi_set(gi_out, pdef_use, grds);

  grds_zgrid_set(GRDS_ZGRID_LEVELS, nz, layer, grds);

  grds_tdef_set(1, *ymdhm, 0, grds);

  grds_opt_list_add("big_endian", grds);
  grds_opt_list_add("sequential", grds);

  grds_var_list_add("TUGD", nz, "soil temperature (K)", grds);
  grds_var_list_add("KIND", 0, "surface kind", grds);

#ifdef DEBUG
    grds_print_var(&grds);
#endif
    return 0;
}

int grds_def_ra(const char *fname_grib2, 
		const grid_info_t *gi_out, const unsigned char pdef_use,
		const ymdhm_t *ymdhm, grds_t *grds){
  
  double layer = 0.0;

  grds_init(grds);

  grds_meta_set(fname_grib2, GRDS_META_DSET, grds);
  grds_meta_set("JMA Precipitation Analysis converted from GRIB2", 
		GRDS_META_TITLE, grds);
  
  grds_undef_set(GRDS_UNDEF, grds);
  grds_gi_set(gi_out, pdef_use, grds);

  grds_zgrid_set(GRDS_ZGRID_LEVELS, 1, &layer, grds);

  grds_tdef_set(1, *ymdhm, 0, grds);

  grds_opt_list_add("big_endian", grds);
  grds_opt_list_add("sequential", grds);

  grds_var_list_add("RAIN", 0, "1-hour accumulated precipitation (mm)", grds);

#ifdef DEBUG
    grds_print_var(&grds);
#endif
    return 0;
}

int grds_def_latlon(const char *fname_grib2, 
		    const grid_info_t *gi_out, const unsigned char pdef_use,
		    const ymdhm_t *ymdhm, grds_t *grds){
  
  double layer = 0.0;
  char fname[256];

  strcpy(fname, fname_grib2);
  strcat(fname, ".latlon");

  grds_init(grds);

  grds_meta_set(fname, GRDS_META_DSET, grds);
  grds_meta_set("LAT and LON", 
		GRDS_META_TITLE, grds);
  
  grds_undef_set(GRDS_UNDEF, grds);
  grds_gi_set(gi_out, pdef_use, grds);
  grds_zgrid_set(GRDS_ZGRID_LEVELS, 1, &layer, grds);

  grds_tdef_set(1, *ymdhm, 0, grds);

  grds_opt_list_add("big_endian", grds);
  grds_opt_list_add("sequential", grds);

  grds_var_list_add("LAT", 0, "Latitude (degree)", grds);
  grds_var_list_add("LON", 0, "Longitude (degree)", grds);

#ifdef DEBUG
    grds_print_var(&grds);
#endif
    return 0;
}

int grds_write_latlon(const grid_info_t *gi,  grds_t *grds){

  ddat_t dlat;
  ddat_t dlon;
  ddat_t dfi;
  ddat_t dfj;
  fdat_t flat;
  fdat_t flon;
  int i;
  int j;

  int r;

  r = 0;
  ddat_init(&dlat);
  ddat_init(&dlon);
  ddat_init(&dfi);
  ddat_init(&dfj);
  fdat_init(&flat);
  fdat_init(&flon);


  ddat_prepare(gi->nxy, &dlat);
  ddat_prepare(gi->nxy, &dlon);
  ddat_prepare(gi->nxy, &dfi);
  ddat_prepare(gi->nxy, &dfj);
  fdat_prepare(gi->nxy, &flat);
  fdat_prepare(gi->nxy, &flon);

  for (j = 0; j < gi->ny; j++){
    for (i = 0; i < gi->nx; i++){
      dfi.data[j * gi->nx + i] = (double)(i + 1);
      dfj.data[j * gi->nx + i] = (double)(j + 1);
    }
  }

  if (memcmp(gi->proj, "LM", 2) == 0){
    NWP_lambert2sphere_D(
			 dfi.data, dfj.data, gi->nxy, 
			 gi->slat1, gi->slat2, gi->slon, 
			 gi->xlat, gi->xlon, gi->xi, gi->xj,
			 gi->dx, dlat.data, dlon.data);
  } else if (memcmp(gi->proj, "LL", 2) == 0){

    for (j = 0; j < gi->ny; j++){
      for (i = 0; i < gi->nx; i++){
	dlon.data[j * gi->nx + i] = gi->xlon 
	  + (dfi.data[j * gi->nx + i] - gi->xi) * gi->dx;
	dlat.data[j * gi->nx + i] = gi->xlat 
	  + (gi->xj - dfj.data[j * gi->nx + i]) * gi->dy;
	if (dlon.data[j * gi->nx + i] < 0.0){
	  dlon.data[j * gi->nx + i] = dlon.data[j * gi->nx + i] + 360.0;
	}
      }
    }

  } else {
    fprintf(stderr, "ERROR! Unknown gridinfo projection\n");
    r =  ERR_UNKNOWN_GRIDINFO_PROJ;
    goto END;
  }

  for (i = 0; i < gi->nxy; i++){
    flat.data[i] = (float)(dlat.data[i]);
    flon.data[i] = (float)(dlon.data[i]);
  }


  grds_dat_write("LAT", 0, &flat, grds);
  grds_dat_write("LON", 0, &flon, grds);

 END:
  ddat_free(&dlat);
  ddat_free(&dlon);
  ddat_free(&dfi);
  ddat_free(&dfj);
  fdat_free(&flat);
  fdat_free(&flon);

  return r;

}

int grds_wrt_map(const char *fname_map_in, const char *fname_map_out, 
		 grid_info_t *gi){
  int r;

  if ((r = gramap(fname_map_in, fname_map_out, gi))) return r;

  return 0;
}
