#include <stdio.h>
#include <stdlib.h>
#include "nwpl_capi.h"
#include "gramap.h"

#define MAXLINE 1024
#define MAXNAME 1024
#define MAXNUM  1024

static FILE	*fp_i,*fp_o;
static double	olat[MAXLINE],olon[MAXLINE];


static int edge(int p1,int p2,int im,int jm){

	if ( olon[p2] < 0. ){
		olat[p2]=(olon[p1]*olat[p2]-olat[p1]*olon[p2])
			/(olon[p1]-olon[p2]);
		olon[p2]=0.;
	}
	if ( olat[p2] < 0. ){
		olon[p2]=(olon[p2]*olat[p1]-olat[p2]*olon[p1])
			/(olat[p1]-olat[p2]);
		olat[p2]=0.;
	}
	if ( olon[p2] > im ){
		olat[p2]=(olon[p1]*olat[p2]-olat[p1]*olon[p2]
			+(olat[p1]-olat[p2])*im)
			/(olon[p1]-olon[p2]);
		olon[p2]=im;
	}
	if ( olat[p2] > jm ){
		olon[p2]=(olon[p2]*olat[p1]-olat[p2]*olon[p1]
			+(olon[p1]-olon[p2])*jm)
			/(olat[p1]-olat[p2]);
		olat[p2]=jm;
	}

	if ( olon[p2] < 0. || olat[p2] < 0. || olon[p2] > im || olat[p2] > jm ){
		edge(p1,p2,im,jm);
	}else{
		return 0;
	}
	return 0;
}

static int ito3b(unsigned long llt,
	unsigned char *l1,unsigned char *l2,unsigned char *l3){
	*l1=(llt & 0x00ff0000)>>16;
	*l2=(llt & 0x0000ff00)>> 8;
	*l3=(llt & 0x000000ff)    ;
	return 0;
}

static int fileout(int imax,unsigned long ltyp,int im,int jm){
	unsigned char	l1,l2,l3;
	unsigned long	llt;
	int		i;
	unsigned long	lonmax,latmax;

	lonmax=im*10000;
	latmax=jm*10000;

	l1=1;	fwrite(&l1,1,1,fp_o);
	l2=ltyp;fwrite(&l2,1,1,fp_o);
	l3=imax;fwrite(&l3,1,1,fp_o);

	for (i=0;i<imax;i++){

		olon[i]=olon[i]*10000.;
		olat[i]=((jm-olat[i]+1)+90.)*10000.;

		llt=olon[i];ito3b(llt,&l1,&l2,&l3);
		fwrite(&l1,1,1,fp_o);
		fwrite(&l2,1,1,fp_o);
		fwrite(&l3,1,1,fp_o);

		llt=olat[i];ito3b(llt,&l1,&l2,&l3);
		fwrite(&l1,1,1,fp_o);
		fwrite(&l2,1,1,fp_o);
		fwrite(&l3,1,1,fp_o);

	}
	return 0;
}

int gramap(const char* fname_i, const char* fname_o, 
	   grid_info_t *gi){

	char		buff[MAXLINE];
	double		**fi,**fj,**flat,**flon;
	int		header,is_in;
	float		rlat1,rlat2,rlat3,rlat4,rlat5;
	float		rlon1,rlon2,rlon3,rlon4,rlon5;
	unsigned int	j,l,lmax,ltyp,omax;
	double		dels, delsx, delsy;
	double          slat,slat2,slon,xi,xj,xlat,xlon;
	int		im,jm;
	char            *npro;

	npro = gi->proj;
        im = gi->nx;
        jm = gi->ny;
        delsx = gi->dx;
        delsy = gi->dy;
        if (delsx != delsy){
            fprintf(stderr, "Not supported for delsx != delsy. Use dels = delsx");
        }
        dels = delsx;
        slat = gi->slat1;
        slat2 = gi->slat2;
        slon = gi->slon;
        xi = gi->xi;
        xj = gi->xj;
        xlat = gi->xlat;
        xlon = gi->xlon;

	if (im*jm == 0){
            fprintf(stderr, "map projection error!\n");
            return -1;
	}

	fp_i = fopen(fname_i,"rt");
	fp_o = fopen(fname_o,"wb");

	if (fp_i == NULL){
            fprintf(stderr, "can't open file error : %s\n", fname_i);
            return -1;
	}

	if (fp_o == NULL){
            fprintf(stderr, "can't open file error : %s\n", fname_o);
            return -1;
	}

	fi   = (double **)(malloc(1 * sizeof(double *)));
	fj   = (double **)(malloc(1 * sizeof(double *)));
	flat = (double **)(malloc(1 * sizeof(double *)));
	flon = (double **)(malloc(1 * sizeof(double *)));

	for(j = 0; j < 1; j++){
            fi  [j]=(double *)(malloc(MAXLINE*sizeof(double)));
            fj  [j]=(double *)(malloc(MAXLINE*sizeof(double)));
            flat[j]=(double *)(malloc(MAXLINE*sizeof(double)));
            flon[j]=(double *)(malloc(MAXLINE*sizeof(double)));
	}

	header = 1;
	l = 0;lmax = 0;

	while(fgets(buff,MAXLINE,fp_i)!=NULL){

            if (header==1){
                sscanf(buff,"%d%d",&lmax,&ltyp);
                if (ltyp==1)ltyp=0;	/* coast line		*/
                if (ltyp==2)ltyp=2;	/* state boundary	*/
                if (ltyp==3)ltyp=1;	/* country boundary	*/
                if (ltyp==4)ltyp=0;	/* coast line		*/
                if (ltyp==5)ltyp=2;	/* prefecture bundary	*/
                if (ltyp==6)ltyp=2;	/* lake 		*/
                lmax /= 2;
                header=0;
            }else{
                sscanf(buff,"%f%f%f%f%f%f%f%f%f%f"
                       ,&rlat1,&rlon1,&rlat2,&rlon2,&rlat3,&rlon3
                       ,&rlat4,&rlon4,&rlat5,&rlon5);
                flat[0][l+0]=rlat1;flon[0][l+0]=rlon1;
                flat[0][l+1]=rlat2;flon[0][l+1]=rlon2;
                flat[0][l+2]=rlat3;flon[0][l+2]=rlon3;
                flat[0][l+3]=rlat4;flon[0][l+3]=rlon4;
                flat[0][l+4]=rlat5;flon[0][l+4]=rlon5;
                l+=5;;
            }
            
            if (lmax >0 && l>=lmax){
                if (strncmp(npro,"LM",2)==0
                    || strncmp(npro,"lm",2)==0){
                    nwp_sphere2lambert_d(
                        *flat, *flon,lmax,
                        slat, slat2, slon, xlat, xlon, xi, xj,
                        dels, *fi, *fj);
                }
                
                if (strncmp(npro,"PS",2)==0
                    || strncmp(npro,"ps",2)==0){
                    nwp_sphere2polar_d(
                        *flat, *flon, lmax,
                        slat, slon, xlat, xlon, xi, xj, 
                        dels, *fi, *fj);
                }
                
                if (strncmp(npro,"ME",2)==0
                    || strncmp(npro,"me",2)==0
                    || strncmp(npro,"MR",2)==0
                    || strncmp(npro,"mr",2)==0   ){
                    nwp_sphere2mercator_d(
                        *flat, *flon, lmax,
                        slat, xlat, xlon, xi, xj, dels, *fi, *fj);
                }
                
                if (strncmp(npro,"LL",2)==0
                    || strncmp(npro,"ll",2)==0){
                    for (l=0;l<lmax;l++){
                        if (flon[0][l]<0.){
                            fi[0][l]=flon[0][l]+360.;
                        }else{
                            fi[0][l]=flon[0][l];
                        }
                        fj[0][l]=jm-flat[0][l]+1;
                    }
                }
                
                omax=0;
                if ( fi[0][0]>0. && fi[0][0]<im
                     &&  fj[0][0]>0. && fj[0][0]<jm ){
                    olon[omax]=fi[0][0];
                    olat[omax]=fj[0][0];
                    omax++;
                    is_in=1;
                }else{
                    is_in=0;
                }
                
                for (l=1;l<lmax;l++){
                    
                    if ( fi[0][l]>0. && fi[0][l]<im
                         &&  fj[0][l]>0. && fj[0][l]<jm ){
                        
                        if ( is_in==1 ){
                            
                            /* in  -> in  */
                            olon[omax]=fi[0][l];
                            olat[omax]=fj[0][l];
                            omax++;
                            
                        }else{
                            
                            /* out -> in  */
                            olon[omax]=fi[0][l-1];
                            olat[omax]=fj[0][l-1];
                            omax++;
                            olon[omax]=fi[0][l];
                            olat[omax]=fj[0][l];
                            edge (omax,omax-1,im,jm);
                            omax++;
                            
                        }
                        is_in=1;
                        
                        
                    }else{
                        
                        if ( is_in==1 ){
                            
                            /* in  -> out */
                            olon[omax]=fi[0][l];
                            olat[omax]=fj[0][l];
                            edge (omax-1,omax,im,jm);
                            omax++;
                            fileout(omax,ltyp,im,jm);
                            omax=0;
                            
                        }
                        is_in=0;
                        
                    }
                    
                }

                if ( is_in==1){fileout(omax,ltyp,im,jm);}
                l=0;lmax=0;header=1;
                
            }
	}
        
	free  (fi  );
	free  (fj  );
	free  (flat);
	free  (flon);
	fclose(fp_i);
	fclose(fp_o);
	return 0;	
}
