#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "grib2_io.h"


void grib2_init(grib2_t *grib2){
  grib2->rec.list = NULL;
  grib2->rec.last = NULL;

  grib2->data.list = NULL;
  grib2->data.last = NULL;

  return;
}

int grib2_open(const char *fname_in, const char *mode, grib2_t *grib2){
  char fname[256];

  strcpy(fname, fname_in);
  if (mode[0] == 'w'){
    strcat(fname, ".grib2.bin");
  }

  if ((grib2->fp = fopen(fname, mode)) == NULL){
    fprintf(stderr, "Cannot open a GRIB2 file: %s, %s\n", fname, mode);
    return ERR_IO_OPEN;
  }
  grib2->fname = (char*)malloc(strlen(fname) + 1);
  strcpy(grib2->fname, fname);

  fseek(grib2->fp, 0, SEEK_SET);
  return 0;
}

void grib2_close(grib2_t *grib2){
  fclose(grib2->fp);
  return;
}


int grib2_data_set(grib2_t *grib2){
  return grib2_read_info(grib2);
}

void grib2_print_sec0(sec0_t *sec0){
  printf("======== section 0 ========\n");
  printf("GRIB Master Table Number: %u\n", (unsigned int)(sec0->master));
  printf("GRIB Edition Number: %u\n", (unsigned int)(sec0->version));
  printf("Total length of GRIB message in octets: %lld\n", (sec0->length));
  return;
}

void grib2_print_sec1(sec1_t *sec1){
  printf("======== section 1 ========\n");
  printf("Length of section in octets: %u\n", sec1->length);
  printf("Number of Section: %u\n", (unsigned int)(sec1->section));
  printf("Originating/generating centre: %u\n", (unsigned int)(sec1->center));
  printf("Originating/generating sub0-centre: %u\n", (unsigned int)(sec1->subcenter));
  printf("GRIB Master Tables Version Number: %u\n", (unsigned int)(sec1->table_version));
  printf("GRIB Local Tables Version Number: %u\n", (unsigned int)(sec1->region_version));
  printf("Significance of Reference Time: %u\n", (unsigned int)(sec1->vtime_mean));
  printf("Year: %u\n", (unsigned int)(sec1->year));
  printf("Month: %u\n", (unsigned int)(sec1->month));
  printf("Day: %u\n", (unsigned int)(sec1->day));
  printf("Hour: %u\n", (unsigned int)(sec1->hour));
  printf("Minute: %u\n", (unsigned int)(sec1->min));
  printf("Seconde: %u\n", (unsigned int)(sec1->sec));
  printf("Production Status: %u\n", (unsigned int)(sec1->status));
  printf("Type of processed data: %u\n", (unsigned int)(sec1->kind));
  return;
}

void grib2_print_sec2(sec2_t *sec2){
  if (sec2 != NULL){
    printf("======== section 2 ========\n");
    printf("Length of section in octets: %u\n", sec2->length);
    printf("Number of Section: %u\n", (unsigned int)(sec2->section));
  }
  return;
}

void grib2_print_sec3(sec3_t *sec3){
  sec3_proj_30_t *sec3_proj_ln;
  sec3_proj_0_t *sec3_proj_ll;

  printf("======== section 3 ========\n");
  printf("Length of section in octets: %u\n", sec3->length);
  printf("Number of Section: %u\n", (unsigned int)(sec3->section));
  printf("Source of grid definition: %u\n", (unsigned int)(sec3->source));
  printf("Number of data points: %u\n", (unsigned int)(sec3->n_grid));
  printf("Number of octets for optional list of numbers: %u\n", (unsigned int)(sec3->oct_optlist));
  printf("Interpretation of list of numbers: %u\n", (unsigned int)(sec3->list_kind));
  printf("Grid Definition Template Number: %u\n", (unsigned int)(sec3->grid_template));

  if (sec3->grid_template == 30){ /* Lambert */
    sec3_proj_ln = (sec3_proj_30_t*)(sec3->proj);
    printf("Shape of the Earth: %u\n", (unsigned int)(sec3_proj_ln->earth_shape));
    printf("Scale factor of radius of spherical Earth: %u\n", (unsigned int)(sec3_proj_ln->r_earth_unit));
    printf("Scaled value of radius of spherical Earth: %u\n", (unsigned int)(sec3_proj_ln->r_earth));
    printf("Scale factor of major axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ln->a_earth_unit));
    printf("Scaled value of major axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ln->a_earth));
    printf("Scale factor of minor axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ln->b_earth_unit));
    printf("Scale value of minor axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ln->b_earth));
    printf("Nx -- Number of points along a parallel: %u\n", (unsigned int)(sec3_proj_ln->nx));
    printf("Ny -- Number of points along a meridian: %u\n", (unsigned int)(sec3_proj_ln->ny));
    printf("La1 -- Latitude of first grid point: %u\n", (unsigned int)(sec3_proj_ln->la1));
    printf("Lo1 -- Longitude of first grid point: %u\n", (unsigned int)(sec3_proj_ln->lo1));
    printf("Resolution and component flags: %u\n", (unsigned int)(sec3_proj_ln->res));
    printf("LaD -- Latitude where Dx and Dy are specified: %u\n", (unsigned int)(sec3_proj_ln->lad));
    printf("LoV -- Longitude of meridian parallel to y-axis: %u\n", (unsigned int)(sec3_proj_ln->lov));
    printf("Dx -- x-direction grid length: %u\n", (unsigned int)(sec3_proj_ln->dx));
    printf("Dy -- y-direction grid length: %u\n", (unsigned int)(sec3_proj_ln->dy));
    printf("Projection centre flag: %u\n", (unsigned int)(sec3_proj_ln->flag_center));
    printf("Scanning mode: %u\n", (unsigned int)(sec3_proj_ln->scan_mode));
    printf("Latin1 -- first latitude from the pole at which the secant cone cuts the sphere: %u\n", (unsigned int)(sec3_proj_ln->latin1));
    printf("latin2 -- second latitude from the pole at which the secant cone cuts the sphere: %u\n", (unsigned int)(sec3_proj_ln->latin2));
    printf("Latitude of the southern pole of projection: %u\n", (unsigned int)(sec3_proj_ln->proj_sp_lat));
    printf("Longitude of the southern pole of projection: %u\n", (unsigned int)(sec3_proj_ln->proj_sp_lon));
  } else if (sec3->grid_template == 0) { /* latlon */
    sec3_proj_ll = (sec3_proj_0_t*)(sec3->proj);
    printf("Shape of the Earth: %u\n", (unsigned int)(sec3_proj_ll->earth_shape));
    printf("Scale factor of radius of spherical Earth: %u\n", (unsigned int)(sec3_proj_ll->r_earth_unit));
    printf("Scaled values of radius of spherical Earth: %u\n", (unsigned int)(sec3_proj_ll->r_earth));
    printf("Scale factor of major axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ll->a_earth_unit));
    printf("Scaled values of major axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ll->a_earth));
    printf("Scale factor of minor axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ll->b_earth_unit));
    printf("Scaled values of minor axis of oblate spheroid Earth: %u\n", (unsigned int)(sec3_proj_ll->b_earth));
    printf("Ni -- number of points along a parallel: %u\n", (unsigned int)(sec3_proj_ll->nx));
    printf("Nj -- nunmber of points along a meridian: %u\n", (unsigned int)(sec3_proj_ll->ny));
    printf("Basic angle of the initial production domain: %u\n", (unsigned int)(sec3_proj_ll->basic_angle));
    printf("Subdivision of basic angle: %u\n", (unsigned int)(sec3_proj_ll->subdivision));
    printf("La1 -- latitude of first grid point: %u\n", (unsigned int)(sec3_proj_ll->la1));
    printf("Lo1 -- longitude of first grid point: %u\n", (unsigned int)(sec3_proj_ll->lo1));
    printf("Resolution and component flags: %u\n", (unsigned int)(sec3_proj_ll->res));
    printf("La2 -- latitude of last grid point: %u\n", (unsigned int)(sec3_proj_ll->la2));
    printf("Lo2 -- longitude of last grid point: %u\n", (unsigned int)(sec3_proj_ll->lo2));
    printf("Di -- i direction increment: %u\n", (unsigned int)(sec3_proj_ll->dx));
    printf("Dj -- j direction increment: %u\n", (unsigned int)(sec3_proj_ll->dy));
    printf("Scanning Mode: %u\n", (unsigned int)(sec3_proj_ll->scan_mode));
  }
  return;

}

void grib2_print_sec4(sec4_t *sec4){
  sec4_prod_0_t *sec4_prod_0;
  sec4_prod_8_t *sec4_prod_8;
  int i;

  printf("======== section 4 ========\n");
  printf("Length of section in octets: %u\n", sec4->length);
  printf("Number of Section: %u\n", (unsigned int)(sec4->section));
  printf("Number of coordinate values after Template: %u\n", (unsigned int)(sec4->n_grid_after_temp));
  printf("Product Definition Template Number: %u\n", (unsigned int)(sec4->prod_template));

  switch (sec4->prod_template){
  case(0):
    sec4_prod_0 = (sec4_prod_0_t*)(sec4->prod_def);
    printf("Parameter category: %u\n", (unsigned int)(sec4_prod_0->param_category));
    printf("Parameter number: %u\n", (unsigned int)(sec4_prod_0->param_number));
    printf("Type of generating process: %u\n", (unsigned int)(sec4_prod_0->kind_process));
    printf("Background generating process identifier: %u\n", (unsigned int)(sec4_prod_0->bg_process));
    printf("Analysis or forecast generating process identified: %u\n", (unsigned int)(sec4_prod_0->process));
    printf("Hours of observational data cutoff after reference time: %u\n", (unsigned int)(sec4_prod_0->cutoff_time_hour));
    printf("Minutes of observational data cutoff after reference time: %u\n", (unsigned int)(sec4_prod_0->cutoff_time_min));
    printf("Indicator of unit of time range: %u\n", (unsigned int)(sec4_prod_0->term_unit));
    printf("Forecast time: %u\n", (unsigned int)(sec4_prod_0->term));
    printf("Type of first fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer1_kind));
    printf("Scale factor of first fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer1_unit));
    printf("Scaled value of first fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer1));    
    printf("Type of second fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer2_kind));
    printf("Scale factor of second fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer2_unit));
    printf("Scaled value of second fixed surface: %u\n", (unsigned int)(sec4_prod_0->fixed_layer2));    
    break;
  case(8):
  case(50008):
    sec4_prod_8 = (sec4_prod_8_t*)(sec4->prod_def);
    printf("Parameter category: %u\n", (unsigned int)(sec4_prod_8->param_category));
    printf("Parameter number: %u\n", (unsigned int)(sec4_prod_8->param_number));
    printf("Type of generating process: %u\n", (unsigned int)(sec4_prod_8->kind_process));
    printf("Background generating process identifier: %u\n", (unsigned int)(sec4_prod_8->bg_process));
    printf("Analysis or forecast generating process identified: %u\n", (unsigned int)(sec4_prod_8->process));
    printf("Hours of observational data cutoff after reference time: %u\n", (unsigned int)(sec4_prod_8->cutoff_time_hour));
    printf("Minutes of observational data cutoff after reference time: %u\n", (unsigned int)(sec4_prod_8->cutoff_time_min));
    printf("Indicator of unit of time range: %u\n", (unsigned int)(sec4_prod_8->term_unit));
    printf("Forecast time: %u\n", (unsigned int)(sec4_prod_8->term));
    printf("Type of first fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer1_kind));
    printf("Scale factor of first fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer1_unit));
    printf("Scaled value of first fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer1));    
    printf("Type of second fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer2_kind));
    printf("Scale factor of second fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer2_unit));
    printf("Scaled value of second fixed surface: %u\n", (unsigned int)(sec4_prod_8->fixed_layer2));    
    printf("Year -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->eyear));
    printf("Month -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->emonth));
    printf("Day -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->eday));
    printf("Hour -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->ehour));
    printf("Minute -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->emin));
    printf("Second -- Time of end of overall time interval: %u\n", (unsigned int)(sec4_prod_8->esec));
    printf("n -- number of time ranges specifications: %u\n", (unsigned int)(sec4_prod_8->n));
    printf("Total number of data values missing in statistical process: %u\n", (unsigned int)(sec4_prod_8->n_miss));
    for (i = 0; i < sec4_prod_8->n; i++){
      printf("(%d) Statistical process: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->stat_process));
      printf("(%d) Type of time increment between successive fields: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->delt_kind));
      printf("(%d) Indicator of unit of time for time range: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->term_unit));
      printf("(%d) Length of the time range: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->term));
      printf("(%d) Indicator of unit of time for the increment between the successive field: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->delt_unit));
      printf("(%d) Time increment between successive fields: %u\n", 
             i, (unsigned int)(sec4_prod_8->stat[i]->delt));
    }
    break;
  }
  return;

}

void grib2_print_sec5(sec5_t *sec5){
  sec5_rep_0_t *sec5_rep_0;
  sec5_rep_200_t *sec5_rep_200;
  int i;

  printf("======== section 5 ========\n");
  printf("Length of section in octets: %u\n", sec5->length);
  printf("Number of Section: %u\n", (unsigned int)(sec5->section));
  printf("Number of data points: %u\n", (unsigned int)(sec5->n_point));
  printf("Data Representation Template Number: %u\n", (unsigned int)(sec5->rep_template));

  switch(sec5->rep_template){
  case(0):
    sec5_rep_0 = (sec5_rep_0_t*)(sec5->rep);
    printf("Reference Value: %f\n", (sec5_rep_0->reference));
    printf("Binary Scale Factor: %d\n", (int)(sec5_rep_0->binary_unit));
    printf("Decimal Scale Factor: %d\n", (int)(sec5_rep_0->decimal_unit));
    printf("Number of bits: %u\n", (unsigned int)(sec5_rep_0->n_bit));
    printf("Type of origina field values: %u\n", (unsigned int)(sec5_rep_0->kind_value));
    break;
  case(200):
    sec5_rep_200 = (sec5_rep_200_t*)(sec5->rep);
    printf("Number of bits: %u\n", (unsigned int)(sec5_rep_200->n_bit));
    printf("MV -- Maximum value within the levels that used in the packing: %u\n", (unsigned int)(sec5_rep_200->max_level_this));
    printf("MVL -- Maxium value of level (Predefined): %u\n", (unsigned int)(sec5_rep_200->max_level));
    printf("Decimal scale factor of representative value of each level: %u\n", (unsigned int)(sec5_rep_200->unit));
    for (i = 0; i < sec5_rep_200->max_level; i++){
      printf("representative values of level [%d]: %u\n", i, (unsigned int)(sec5_rep_200->lev_to_prec[i]));      
    }
    break;
  }

  return;
}

void grib2_print_sec6(sec6_t *sec6){

  printf("======== section 6 ========\n");
  printf("Length of section in octets: %u\n", sec6->length);
  printf("Number of Section: %u\n", (unsigned int)(sec6->section));
  printf("Bit-map indicator: %u\n", (unsigned int)(sec6->bitmap_flag));
  printf("Position of bit-map in a file: %lld\n", (sec6->bitmap_pos));
  return;
}

void grib2_print_sec7(sec7_t *sec7){
  printf("======== section 7 ========\n");
  printf("Length of section in octets: %u\n", sec7->length);
  printf("Number of sSction: %u\n", (unsigned int)(sec7->section));
  printf("Position of values in a file: %lld\n", (sec7->value_pos));

  return;
}



void grib2_print_rec(grib2_t *grib2){

  grib2_rec_t *rec = &(grib2->rec);
  grib2_rec_list_t *p;

  for (p = rec->list; p!= NULL; p = p->next){
    if (p->section == 0){
      grib2_print_sec0((sec0_t*)(p->rec));
    } else if (p->section == 1){
      grib2_print_sec1((sec1_t*)(p->rec));
    } else if (p->section == 2){
      grib2_print_sec2((sec2_t*)(p->rec));
    } else if (p->section == 3){
      grib2_print_sec3((sec3_t*)(p->rec));
    } else if (p->section == 4){
      grib2_print_sec4((sec4_t*)(p->rec));
    } else if (p->section == 5){
      grib2_print_sec5((sec5_t*)(p->rec));
    } else if (p->section == 6){
      grib2_print_sec6((sec6_t*)(p->rec));
    } else if (p->section == 7){
      grib2_print_sec7((sec7_t*)(p->rec));
    }

  }
  return;
}

void grib2_print_data(grib2_t *grib2){
  grib2_data_t *data = &(grib2->data);
  grib2_data_list_t *p;

  for (p = data->list; p!= NULL; p = p->next){
    grib2_print_sec0(p->sec0);
    grib2_print_sec1(p->sec1);
    grib2_print_sec2(p->sec2);
    grib2_print_sec3(p->sec3);
    grib2_print_sec4(p->sec4);
    grib2_print_sec5(p->sec5);
    grib2_print_sec6(p->sec6);
    grib2_print_sec7(p->sec7);
  }

  return;
}







void grib2_ymdhm_set_from_sec1(const sec1_t *sec1, ymdhm_t *ymdhm){

  ymdhm->year = sec1->year;
  ymdhm->month = sec1->month;
  ymdhm->day = sec1->day;
  ymdhm->hour = sec1->hour;
  ymdhm->min = sec1->min;

  return;
}


float grib2_inq_earth_rad(const sec3_t *sec3){
  float rad;

  sec3_proj_0_t *sec3_proj_0;
  sec3_proj_30_t *sec3_proj_30;

  switch(sec3->grid_template){
  case (0):  /* lat-lon */
    sec3_proj_0 = (sec3_proj_0_t*)(sec3->proj);
    if (sec3_proj_0->r_earth_unit != 0xff){
      rad = sec3_proj_0->r_earth * pow(10.0, -sec3_proj_0->r_earth_unit);
    } else if (sec3_proj_0->earth_shape == EARTH_SHAPE_GRS80){
      rad = NWP_EARTH_GRS80_RAD_M;
    } else {
      rad = FLT_MAX;
    }
    break;
  case (30):  /* Lambert */
    sec3_proj_30 = (sec3_proj_30_t*)(sec3->proj);
    if (sec3_proj_30->r_earth_unit != 0xff){
      rad = sec3_proj_30->r_earth * pow(10.0, -sec3_proj_30->r_earth_unit);
    } else if (sec3_proj_30->earth_shape == EARTH_SHAPE_GRS80){
      rad = NWP_EARTH_GRS80_RAD_M;      
    } else {
      rad = FLT_MAX;
    }
    break;
  }
  return rad;
}


int grib2_get_jma_kind(grib2_t *grib2){
  int r;
  grib2_data_list_t *data_list = grib2->data.list;

  /* look at the first data */
  if (data_list->sec0->master == 0){
    if (data_list->sec4->prod_template == 0 
	|| data_list->sec4->prod_template == 8){
      r = JMA_KIND_ATM;
    } else if (data_list->sec4->prod_template == 50008){
      r = JMA_KIND_RA;
    } else {
      fprintf(stderr, "unknown GRIB2 file for meteorology.\n");
      r = -1;
    }
  } else if (data_list->sec0->master == 2){
    r = JMA_KIND_LAND;
  } else if (data_list->sec0->master == 10){
    r = JMA_KIND_OCEAN;
  } else{
      fprintf(stderr, "unknown GRIB2 master.\n");
      r = -2;
  }
  return r;
}
