#ifndef GRIB2_IO_H_INCLUDED
#define GRIB2_IO_H_INCLUDED
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include "typedef.h"
#include "ymdhm.h"
#include "codec.h"
#include "nwpl_map_earth.h"

typedef struct sec0_t{
    N_UI1 master;
    N_UI1 version;
    N_UI8 length;
} sec0_t;

typedef struct sec1_t{
    N_UI4 length;
    N_UI1 section;
    N_UI2 center;
    N_UI2 subcenter;
    N_UI1 table_version;
    N_UI1 region_version;
    N_UI1 vtime_mean;
    N_UI2 year;
    N_UI1 month;
    N_UI1 day;
    N_UI1 hour;
    N_UI1 min;
    N_UI1 sec;
    N_UI1 status;
    N_UI1 kind;
} sec1_t;

typedef struct sec2_t{
    N_UI4 length;
    N_UI1 section;
} sec2_t;

typedef struct sec3_t{
    N_UI4 length;
    N_UI1 section;
    N_UI1 source;
    N_UI4 n_grid;
    N_UI1 oct_optlist;
    N_UI1 list_kind;
    N_UI2 grid_template;
    void *proj;
} sec3_t;


typedef struct sec3_proj_30_t{ /* Lambert */
    N_UI1 earth_shape;
    N_UI1 r_earth_unit;
    N_UI4 r_earth;
    N_UI1 a_earth_unit;
    N_UI4 a_earth;
    N_UI1 b_earth_unit;
    N_UI4 b_earth;
    N_UI4 nx;
    N_UI4 ny;
    N_UI4 la1;
    N_UI4 lo1;
    N_UI1 res;
    N_UI4 lad;
    N_UI4 lov;
    N_UI4 dx;
    N_UI4 dy;
    N_UI1 flag_center;
    N_UI1 scan_mode;
    N_UI4 latin1;
    N_UI4 latin2;
    N_UI4 proj_sp_lat;
    N_UI4 proj_sp_lon;
} sec3_proj_30_t;

typedef struct sec3_proj_0_t{ /* Lat-Lon */
    N_UI1 earth_shape;
    N_UI1 r_earth_unit;
    N_UI4 r_earth;
    N_UI1 a_earth_unit;
    N_UI4 a_earth;
    N_UI1 b_earth_unit;
    N_UI4 b_earth;
    N_UI4 nx;
    N_UI4 ny;
    N_UI4 basic_angle;
    N_UI4 subdivision;
    N_UI4 la1;
    N_UI4 lo1;
    N_UI1 res;
    N_UI4 la2;
    N_UI4 lo2;
    N_UI4 dx;
    N_UI4 dy;
    N_UI1 scan_mode;
} sec3_proj_0_t;

typedef struct sec4_t{
    N_UI4 length;
    N_UI1 section;
    N_UI2 n_grid_after_temp;
    N_UI2 prod_template;
    void *prod_def;
} sec4_t;

typedef struct sec4_prod_0_t{
    N_UI1 param_category;
    N_UI1 param_number;
    N_UI1 kind_process;
    N_UI1 bg_process;
    N_UI1 process;
    N_UI2 cutoff_time_hour;
    N_UI1 cutoff_time_min;
    N_UI1 term_unit;
    N_SI4 term;
    N_UI1 fixed_layer1_kind;
    N_UI1 fixed_layer1_unit;
    N_UI4 fixed_layer1;
    N_UI1 fixed_layer2_kind;
    N_UI1 fixed_layer2_unit;
    N_UI4 fixed_layer2;
} sec4_prod_0_t;

typedef struct sec4_prod_8_stat_t{
  N_UI1 stat_process;
  N_UI1 delt_kind;
  N_UI1 term_unit;
  N_SI4 term;
  N_UI1 delt_unit;
  N_UI4 delt;
} sec4_prod_8_stat_t;


typedef struct sec4_prod_8_t{
  N_UI1 param_category;
  N_UI1 param_number;
  N_UI1 kind_process;
  N_UI1 bg_process;
  N_UI1 process;
  N_UI2 cutoff_time_hour;
  N_UI1 cutoff_time_min;
  N_UI1 term_unit;
  N_SI4 term;
  N_UI1 fixed_layer1_kind;
  N_UI1 fixed_layer1_unit;
  N_UI4 fixed_layer1;
  N_UI1 fixed_layer2_kind;
  N_UI1 fixed_layer2_unit;
  N_UI4 fixed_layer2;
  N_UI2 eyear;
  N_UI1 emonth;
  N_UI1 eday;
  N_UI1 ehour;
  N_UI1 emin;
  N_UI1 esec;
  N_UI1 n;
  N_UI4 n_miss;
  sec4_prod_8_stat_t **stat;

  /* only for 50008 */
  N_UI8 radar_use1;
  N_UI8 radar_use2;
  N_UI8 raingauge_use;

} sec4_prod_8_t;



typedef struct sec5_t{
    N_UI4 length;
    N_UI1 section;
    N_UI4 n_point;
    N_UI2 rep_template;
    void *rep;
} sec5_t;

typedef struct sec5_rep_0_t{
  float reference;
  N_SI2 binary_unit;
  N_SI2 decimal_unit;
  N_UI1 n_bit;
  N_UI1 kind_value;
} sec5_rep_0_t;

typedef struct sec5_rep_200_t{
  N_UI1 n_bit;
  N_UI2 max_level_this;
  N_UI2 max_level;
  N_UI1 unit;
  N_UI2 *lev_to_prec;
} sec5_rep_200_t;

typedef struct sec6_t{
  N_UI4 length;
  N_UI1 section;
  N_UI1 bitmap_flag;
  N_UI8 bitmap_pos; /* use only in decode */
  N_UI1 *bitmap;    /* use only in encode */
} sec6_t;

typedef struct sec7_t{
  N_UI4 length;
  N_UI1 section;
  N_UI8 value_pos;   /* use only in decode */
  N_UI1 *value;      /* use only in encode */
} sec7_t;



typedef struct grib2_rec_list_t grib2_rec_list_t;


struct grib2_rec_list_t{
  N_UI4 length;
  N_UI1 section;
  void *rec;
  grib2_rec_list_t *next;
};


/* elements should be initialized with NULL */
typedef struct grib2_rec_t{
  grib2_rec_list_t *list;
  grib2_rec_list_t *last;
} grib2_rec_t;

typedef struct grib2_data_list_t grib2_data_list_t;

struct grib2_data_list_t{
  sec0_t *sec0;
  sec1_t *sec1;
  sec2_t *sec2;
  sec3_t *sec3;
  sec4_t *sec4;
  sec5_t *sec5;
  sec6_t *sec6;
  sec7_t *sec7;
  grib2_data_list_t *next;
};

/* elements should be initialized with NULL */
typedef struct grib2_data_t{
  grib2_data_list_t *list;
  grib2_data_list_t *last;
} grib2_data_t;

typedef struct grib2_t{
  char *fname;
  FILE *fp;
  grib2_rec_t rec;
  grib2_data_t data;
} grib2_t;


/* prototype */
void grib2_init(grib2_t *grib2);
int grib2_open(const char *fname, const char *mode, grib2_t *grib2);
void grib2_close(grib2_t *grib2);
int grib2_data_set(grib2_t *grib2);
void grib2_print_rec(grib2_t *grib2);
void grib2_print_data(grib2_t *grib2);

void grib2_ymdhm_set_from_sec1(const sec1_t *sec1, ymdhm_t *ymdhm);

float grib2_inq_earth_rad(const sec3_t *sec3);

int grib2_get_jma_kind(grib2_t *grib2);

void grib2_print_sec0(sec0_t *sec1);
void grib2_print_sec1(sec1_t *sec1);
void grib2_print_sec2(sec2_t *sec2);
void grib2_print_sec3(sec3_t *sec3);
void grib2_print_sec4(sec4_t *sec4);
void grib2_print_sec5(sec5_t *sec5);
void grib2_print_sec6(sec6_t *sec6);
void grib2_print_sec7(sec7_t *sec7);


/* grib2_decode.c */
int grib2_inq_nz(grib2_t *grib2, ymdhm_t *ymdhm, N_SI4 term, 
		 N_UI1 parm_cat, N_UI1 parm_num);

int grib2_read_info(grib2_t *grib2);

int grib2_read(grib2_t *grib2,
	       const ymdhm_t *ymdhm, const N_SI4 term, 
	       const N_UI1 parm_cat, const N_UI1 parm_num, 
	       const N_UI1 layer_kind, const N_UI4 nz, 
	       const double *layer, float *data_out);


#endif
