#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "grid_info.h"

void grid_info_set_from_config_h(const config_h_t *config_h, grid_info_t *gi){
  strcpy(gi->proj, config_h->proj);
  gi->nx = config_h->nx;
  gi->ny = config_h->ny;
  gi->dx = config_h->dx;
  gi->dy = config_h->dy;
  gi->xi = config_h->xi;
  gi->xj = config_h->xj;
  gi->xlat = config_h->xlat;
  gi->xlon = config_h->xlon;
  gi->nxy = gi->nx * gi->ny;

  /* derived */
  if (memcmp(gi->proj, "LL", 2) == 0){
    gi->slat1 = 0.0;
    gi->slat2 = 0.0;
    gi->slon = 0.0;

  } else if (memcmp(gi->proj, "LM", 2) == 0){
    /*
    gi->slat1 = config_h->slat1;
    gi->slat2 = config_h->slat2;
    */
    /* force to set */
    gi->slat1 = 30.0;
    gi->slat2 = 60.0;
    gi->slon = config_h->slon;

  }


  return;
}

void grid_info_set_from_sec3(sec3_t *sec3, grid_info_t *gi){

  sec3_proj_30_t *proj_30;
  sec3_proj_0_t *proj_0;
  double lat;
  double lon;
  double fi;
  double fj;

  switch (sec3->grid_template){
  case (0): /* lat-lon */
    memcpy(gi->proj, "LL  ", 2);

    proj_0 = (sec3_proj_0_t*)(sec3->proj);
    gi->nx = proj_0->nx;
    gi->ny = proj_0->ny;
    gi->nxy = gi->nx * gi->ny;

    gi->dx = proj_0->dx * 1.0e-6;
    gi->dy = proj_0->dy * 1.0e-6;

    gi->slat1 = 0.0;  /* dummy */
    gi->slat2 = 0.0;  /* dummy */
    gi->slon = 0.0;   /* dummy */

    gi->xi = 1.0;     /* dummy */
    gi->xj = 1.0;     /* dummy */
    gi->xlat = proj_0->la1 * 1.0e-6;
    gi->xlon = proj_0->lo1 * 1.0e-6;

    break;
  case (30): /* Lambert */
    memcpy(gi->proj, "LMN ", 2);
    
    proj_30 = (sec3_proj_30_t*)(sec3->proj);
    gi->nx = proj_30->nx;
    gi->ny = proj_30->ny;
    gi->nxy = gi->nx * gi->ny;

    gi->dx = proj_30->dx * 1.0e-3;
    gi->dy = proj_30->dy * 1.0e-3;

    gi->slat1 = proj_30->latin2 * 1.0e-6;
    gi->slat2 = proj_30->latin1 * 1.0e-6;
    gi->slon = proj_30->lov * 1.0e-6;

    gi->xi = 1.0;
    gi->xj = 1.0;
    gi->xlat = proj_30->la1 * 1.0e-6;
    gi->xlon = proj_30->lo1 * 1.0e-6;

    lat = proj_30->lad * 1.0e-6;
    lon = proj_30->lov * 1.0e-6;

    NWP_sphere2lambert_D(&lat, &lon, 
                         1, gi->slat1, gi->slat2, 
                         gi->slon, gi->xlat, gi->xlon, 
                         gi->xi, gi->xj, gi->dx,
                         &fi, &fj);
    /* update */
    gi->xi = fi;
    gi->xj = fj;
    gi->xlat = lat;
    gi->xlon = lon;

    break;

  }
  return;

}


void grid_info_print(grid_info_t *gi){
  fprintf(stderr, "nx: %d\n", gi->nx);
  fprintf(stderr, "ny: %d\n", gi->ny);
  fprintf(stderr, "nxy: %d\n", gi->nxy);
  fprintf(stderr, "dx: %f\n", gi->dx);
  fprintf(stderr, "dy: %f\n", gi->dy);
  fprintf(stderr, "xi: %f\n", gi->xi);
  fprintf(stderr, "xj: %f\n", gi->xj);
  fprintf(stderr, "xlat: %f\n", gi->xlat);
  fprintf(stderr, "xlon: %f\n", gi->xlon);
  fprintf(stderr, "slat1: %f\n", gi->slat1);
  fprintf(stderr, "slat2: %f\n", gi->slat2);
  fprintf(stderr, "slon: %f\n", gi->slon);
  return;
}
