#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "parse_config.h"

#define BUFSIZE 1024
#define PNUM_MAX 300

#define CSV_ITEM_MAX 100

extern int errno;

static void parse_config_each(const char *label_str, const int var_kind,
			     unsigned int *flag_arg,
			     const unsigned int ic, char **pp, 
			     const unsigned int n_ary, 
			     const int bit, unsigned char *p_str){
  unsigned int itmp;
  double dtmp;
  unsigned int len;
  unsigned int ii;
  unsigned int i_end;
  unsigned int flag;

  flag = *flag_arg;
  if ((flag & (1<<bit)) == 1){
    fprintf(stderr, "A duplicated entry in %s.\n", label_str);	
  } else {
    i_end = ic < n_ary ? ic : n_ary;
    switch (var_kind){
    case (KIND_I4):
      for (ii = 0; ii < i_end; ii++){
	itmp = strtol(pp[ii], NULL, 10);
	if (errno == 0){
	  memcpy(p_str + ii * 4, (unsigned char*)&itmp, 4);
	}
      }
      break;
    case (KIND_R4):
      for (ii = 0; ii < i_end; ii++){
	dtmp = strtod(pp[ii], NULL);
	if (errno == 0){
	  memcpy(p_str + ii * 4, (unsigned char*)&dtmp, sizeof(double));
	}
      }
      break;
    case (KIND_R8):
      for (ii = 0; ii < i_end; ii++){
	dtmp = strtod(pp[ii], NULL);
	if (errno == 0){
	  memcpy(p_str + ii * sizeof(double),
		 (unsigned char*)&dtmp, sizeof(double));
	}
      }
      break;
    case (KIND_PROJ):  /* only 1 item */
      len = strlen(pp[0]);
      if (len > 4){
	errno = 1;
      } else {
	memcpy(p_str, pp[0], len);
	for (ii = len; ii < 4; ii++){
	  p_str[ii] = ' ';
	}
	p_str[4] = '\0';
      }
      break;
    }

    if (errno){
      fprintf(stderr, "Invalid format in %s.\n", label_str);
    } else {
      flag |= (1<<bit);
    }
  }
  *flag_arg = flag;
  return;
}

static void parse_csv(char *buf, unsigned int *ic, char **pp, char **label_str){
    char *p;
    char *p_eq;
    char *val_str;
    int i;
    int j;

    buf[strlen(buf) - 1] = 0;
    p = buf;
    p_eq = NULL;
    while(*p){
      if (*p == '#'){
	*p = 0;
	break;
      }
      p++;
    }
    if (strlen(buf) == 0) return;
    p = buf;
    /* remove white spaces before labels */
    while(*p){
      if (*p != ' '){
	break;
      }
      p++;
    }
    *label_str = p;

    /* search '=' */
    p = *label_str;
    while (*p){
      if (*p == '='){
	p_eq = p;
	break;
      }
      p++;
    }
    if (*p == 0) return;  /* not found '=' */
    /* replace '=' with \0 */
    *p = 0;
    p--;

    /* remove white spaces after labels */
    while (*p == ' '){
      *p = 0;
      p--;
    }

    /* get values */
    p = p_eq + 1;
    val_str = NULL;

    /* remove white space before values */
    while(*p){
      if (*p != ' '){
	val_str = p;
	break;
      }
      p++;
    }

    /* split CSV in values */
    *ic = 0;
    pp[(*ic)++] = val_str;
    p = val_str;
    while(*p){
      if (*p == ','){
	*p = 0;
	pp[(*ic)++] = ++p;
      } else {
	p++;
      }
    }

    /* remove white spaces before and after each value */
    for (i = 0; i < *ic; i++){
      while (*(pp[i]) == ' '){
	pp[i]++;
      }

      p = pp[i] + strlen(pp[i]) - 1;
      while (*p == ' '){
	*p = 0;
	p--;
      }
    }
    
    /* remove elems with strlen == 0 */
    for (i = 0; i < *ic; i++){
      if (*pp[i] == 0){
	for (j = i; j < *ic - 1; j++){
	  pp[j] = pp[j + 1];
	}
	(*ic)--;
      }
    }
    return;
}

int parse_config_h(const char *fname, config_h_t *config_h){
  FILE *fp;
  char buf[BUFSIZE];

  int r;
  char *label_str;
  char **pp;
  unsigned int ic;
  unsigned int flag;
  unsigned int n_ary;

  if ((pp = (char**)malloc(CSV_ITEM_MAX * sizeof(char*))) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  flag = 0;
  if ((fp = fopen(fname, "r")) == NULL){
    fprintf(stderr, "Cannot open the file: %s.\n", fname);
    return ERR_IO_OPEN;
  }
  n_ary = 1;
  while(fgets(buf, BUFSIZE, fp)){
    parse_csv(buf, &ic, pp, &label_str);

    if (strcmp(label_str, "proj") == 0){
      parse_config_each(label_str, KIND_PROJ, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_PROJ, (unsigned char*)(config_h->proj));
    } else if (strcmp(label_str, "nx") == 0){
      parse_config_each(label_str, KIND_I4, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_NX, (unsigned char*)&(config_h->nx));
    } else if (strcmp(label_str, "ny") == 0){
      parse_config_each(label_str, KIND_I4, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_NY, (unsigned char*)&(config_h->ny));
    } else if (strcmp(label_str, "dx") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_DX, (unsigned char*)&(config_h->dx));
    } else if (strcmp(label_str, "dy") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_DY, (unsigned char*)&(config_h->dy));
    } else if (strcmp(label_str, "xlat") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_XLAT, (unsigned char*)&(config_h->xlat));
    } else if (strcmp(label_str, "xlon") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_XLON, (unsigned char*)&(config_h->xlon));
    } else if (strcmp(label_str, "xi") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_XI, (unsigned char*)&(config_h->xi));
    } else if (strcmp(label_str, "xj") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_XJ, (unsigned char*)&(config_h->xj));
    } else if (strcmp(label_str, "slat1") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_SLAT1, (unsigned char*)&(config_h->slat1));
    } else if (strcmp(label_str, "slat2") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_SLAT2, (unsigned char*)&(config_h->slat2));
    } else if (strcmp(label_str, "slon") == 0){
      parse_config_each(label_str, KIND_R8, &flag,
			ic, pp, n_ary, 
			SHIFT_BIT_SLON, (unsigned char*)&(config_h->slon));
    }
  }

  free(pp);
  fclose(fp);

  r = 0;

  if ((flag & (1 << SHIFT_BIT_PROJ)) == 0){
    fprintf(stderr, "proj is not set in config_h.\n");
    r += 1;
  } else if ((memcmp(config_h->proj, "LL", 2) != 0 
	      && memcmp(config_h->proj, "LM", 2) != 0)){
    fprintf(stderr, "Not supported projection specified in config_h: %s.\n", 
	    config_h->proj);
    r += 1;
  } else {
    if ((flag & (1 << SHIFT_BIT_NX)) == 0){
      fprintf(stderr, "nx is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_NY)) == 0){
      fprintf(stderr, "ny is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_DX)) == 0){
      fprintf(stderr, "dx is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_DY)) == 0){
      fprintf(stderr, "dy is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_XLAT)) == 0){
      fprintf(stderr, "xlat is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_XLON)) == 0){
      fprintf(stderr, "xlon is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_XI)) == 0){
      fprintf(stderr, "xi is not set in config_h.\n");      
      r += 1;
    }
    if ((flag & (1 << SHIFT_BIT_XJ)) == 0){
      fprintf(stderr, "xj is not set in config_h.\n");      
      r += 1;
    }
    if (memcmp(config_h->proj, "LL", 2) == 0){
    } else if (memcmp(config_h->proj, "LM", 2) == 0){
      if ((flag & (1 << SHIFT_BIT_SLON)) == 0){
        fprintf(stderr, "slon is not set in config_h.\n");      
        r += 1;
      }
    }
  }
  if (r) r = ERR_CONFIG_HV_INSUFFICIENT;
  return r;
}

int parse_config_v(const char *fname, config_v_t *config_v){

  unsigned int flag;
  FILE *fp;
  char buf[BUFSIZE];
  unsigned int ic;
  double *pout_tmp;
  unsigned int pout_num;
  double dtmp;
  char **pp;
  unsigned int n_ary;
  char *label_str;
  unsigned int ii;

  if ((pp = (char**)malloc(CSV_ITEM_MAX * sizeof(char*))) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  flag = 0;
  if ((fp = fopen(fname, "r")) == NULL){
    fprintf(stderr, "Cannot open the file: %s.\n", fname);
    return ERR_IO_OPEN;
  }
  flag = 0;
  n_ary = PNUM_MAX;
  
  pout_tmp = (double*)malloc(PNUM_MAX * sizeof(double));
  pout_num = 0;

  while(fgets(buf, BUFSIZE, fp)){
    parse_csv(buf, &ic, pp, &label_str);
    if (strcmp(label_str, "pout") == 0){
      config_v->pnum = 0;
      for (ii = 0; ii < ic; ii++){
	dtmp = strtod(pp[ii], NULL);
	if (errno){
	} else{
	  pout_tmp[pout_num++] = dtmp;
	}

      }
      config_v->pnum = pout_num;
      config_v->pout = (double*)malloc(pout_num * sizeof(double));
      memcpy(config_v->pout, pout_tmp, pout_num * sizeof(double));
    }
  }
  fclose(fp);
  if (config_v->pnum == 0){
    fprintf(stderr, "pout is not correctly specified in config_v.\n");
    return ERR_CONFIG_HV_INSUFFICIENT;
  }

  return 0;
}

void parse_config_print_config_h(const config_h_t *config_h){
  fprintf(stderr, "=== Projection information obtained from config_h file ===\n");
  fprintf(stderr, "proj: %s\n", config_h->proj);
  fprintf(stderr, "nx: %d\n", config_h->nx);
  fprintf(stderr, "ny: %d\n", config_h->ny);
  fprintf(stderr, "dx: %f\n", config_h->dx);
  fprintf(stderr, "dy: %f\n", config_h->dy);
  fprintf(stderr, "xlat: %f\n", config_h->xlat);
  fprintf(stderr, "xlon: %f\n", config_h->xlon);
  fprintf(stderr, "xi: %f\n", config_h->xi);
  fprintf(stderr, "xj: %f\n", config_h->xj);
  if (memcmp(config_h->proj, "LM", 2) == 0){
    fprintf(stderr, "slat1: %f\n", config_h->slat1);
    fprintf(stderr, "slat2: %f\n", config_h->slat2);
    fprintf(stderr, "slon: %f\n", config_h->slon);
  }
  return ;
}

void parse_config_print_config_v(const config_v_t *config_v){
  int i;
  fprintf(stderr, "=== Vertical isobaric planes obtained from config_v file ===\n");
  fprintf(stderr, "pnum: %d\n", config_v->pnum);
  
  for (i = 0; i < config_v->pnum; i++){
    fprintf(stderr, "pout[%d]: %f\n", i, config_v->pout[i]);
  }
  return;
}

