#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "parse_opt.h"


int parse_opt(int argc, char **argv, opt_t *opt){
  int i;

  opt->fname_config_h = NULL;
  opt->fname_config_v = NULL;
  opt->fname_grib2 = NULL;
  opt->fname_out = NULL;
  opt->print_rec = 0;
  opt->allow_out_domain = 0;
  opt->ra_interp_ope = INTERP_OPE_MEAN;
  opt->output_format = FORMAT_GRADS;
  opt->output_latlon = 0;
  opt->fname_mapfile_in = NULL;
  opt->pdef_use = 0;
  opt->udg_miss = 0;

  for (i = 1; i < argc; i++){
    if (*argv[i] == '-'){
      switch (*(argv[i] + 1)){
      case 'h':
	i++;
	if (i == argc || argv[i][0] == '-'){
	  fprintf(stderr, "No config file is specified after -c.\n");
	  return ERR_OPT_NOFILE_SPECIFIED;
	}
	if (opt->fname_config_h == NULL){
	  opt->fname_config_h = argv[i];
	} else {
	  fprintf(stderr, "More than one config file are specified.\n");
	  return ERR_OPT_DUP_FILE;
	}
	break;
      case 'v':
	i++;
	if (i == argc || argv[i][0] == '-'){
	  fprintf(stderr, "No config file is specified after -c.\n");
	  return ERR_OPT_NOFILE_SPECIFIED;
	}
	if (opt->fname_config_v == NULL){
	  opt->fname_config_v = argv[i];
	} else {
	  fprintf(stderr, "More than one config file are specified.\n");
	  return ERR_OPT_DUP_FILE;
	}
	break;
      case 'o':
	i++;
	if (i == argc || argv[i][0] == '-'){
	  fprintf(stderr, "No output file is specified after -o.\n");
	  return ERR_OPT_NOFILE_SPECIFIED;
	}
	if (opt->fname_out == NULL){
	  opt->fname_out = argv[i];
	} else {
	  fprintf(stderr, "More than one output file are specified.\n");
	  return ERR_OPT_DUP_FILE;
	}
	break;
      case 'c':
	opt->pdef_use = 1;
	break;
      case 'p':
	opt->print_rec = 1;
	break;
      case 'd':
	opt->allow_out_domain = 1;
	break;
      case 'g':
	opt->output_format = FORMAT_GRIB2;
	break;
      case 'l':
	opt->output_latlon = 1;
	break;
      case 'u':
	opt->udg_miss = 1;
	break;
      case 'm':
	i++;
	if (i == argc || argv[i][0] == '-'){
	  fprintf(stderr, "No map file is not specified after -m.\n");
	  return ERR_OPT_NOFILE_SPECIFIED;
	}
	if (opt->fname_mapfile_in == NULL){
	  opt->fname_mapfile_in = argv[i];
	} else {
	  fprintf(stderr, "More than one mapfile are specified.\n");
	  return ERR_OPT_DUP_FILE;
	}
	break;
      case 'r':
	i++;
	if (i == argc || argv[i][0] == '-'){
	  fprintf(stderr, "No identifier for RA interpolation option after -r.\n");
	  return ERR_OPT_NOFILE_SPECIFIED;
	}
	switch (*(argv[i])){
	case 'm':
	  opt->ra_interp_ope = INTERP_OPE_MEAN;	  
	  break;
	case 'n':
	  opt->ra_interp_ope = INTERP_OPE_NEAR;
	  break;
	case 'x':
	  opt->ra_interp_ope = INTERP_OPE_MAX;
	  break;
	default:
	  fprintf(stderr, "the interpolation method for RA is invalid: %s\n", 
		  argv[i]);
	  return ERR_OPT_INVARID;
	}
	break;
      }
    } else {
      if (opt->fname_grib2 == NULL){
	opt->fname_grib2 = argv[i];
      } else {
	fprintf(stderr, "Duplicated input GRIB2 files.\n");
	return ERR_OPT_DUP_FILE;
      }
    }
  }
  if (opt->fname_grib2 == NULL){
    fprintf(stderr, "A GRIB2 file is not specified.\n");
    return ERR_OPT_NOFILE_SPECIFIED;
  }
  if (opt->fname_out == NULL){
    if ((opt->fname_out = (char*)malloc(strlen(opt->fname_grib2) + 1)) == NULL){
      fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
      return ERR_MALLOC;
    }
    strcpy(opt->fname_out, opt->fname_grib2);
  }
  return 0;
}


void parse_opt_print(opt_t *opt){
  fprintf(stderr, "=== Specified options ===\n");
  fprintf(stderr, "fname_grib2: %s\n", opt->fname_grib2);
  if (opt->fname_config_h != NULL){
    fprintf(stderr, "fname_config_h: %s\n", opt->fname_config_h);    
  } else {
    fprintf(stderr, "fname_config_h: not specified\n");        
  }
  if (opt->fname_config_v != NULL){
    fprintf(stderr, "fname_config_v: %s\n", opt->fname_config_v);    
  } else {
    fprintf(stderr, "fname_config_v: not specified\n");        
  }
  fprintf(stderr, "fname_out: %s\n", opt->fname_out);

  return;
}

void parse_opt_usage(char *cmd){
  fprintf(stderr, "usage: %s grib2_file [-h config_h_file] [-v config_v_file] [-o output_file] [-g] [-p] [-d] [-r m|x|n] [-l] [-c] [-u]\n", cmd);
  fprintf(stderr, "      -g: output in GRIB2 format\n");
  fprintf(stderr, "      -p: only print records in GRIB2\n");
  fprintf(stderr, "      -d: allow out of domain in coordinate conversion\n");
  fprintf(stderr, "      -c: use pdef in GrADS ctl files\n");
  fprintf(stderr, "      -u: set MISSING to values located undergroud\n");
  fprintf(stderr, "      -r: RA interpolation option\n");
  fprintf(stderr, "          m: mean\n");
  fprintf(stderr, "          x: max\n");
  fprintf(stderr, "          n: nearest\n");
  fprintf(stderr, "      -l: output lat and lon in GrADS format\n");
  return;
}
